# -*- coding: utf-8 -*-
"""
Submodule for INS sensor models.

This module generates synthetic sensor data of a hypothetical strapdown
inertial navigation system (INS.) It mimics a 9-DOF IMU (3-axes gyroscope,
3-axes accelerometer, and 3-axes magnetometer) from a given array of
orientations as quaternions.

"""

import numpy as np
from ..common.constants import DEG2RAD
from ..common.constants import RAD2DEG
from ..common.constants import MUNICH_LATITUDE
from ..common.constants import MUNICH_LONGITUDE
from ..common.constants import MUNICH_HEIGHT
from ..common.frames import ned2enu
from ..common.quaternion import QuaternionArray
from ..utils import WMM
from ..utils import WGS

GENERATOR = np.random.default_rng(42)

# Geomagnetic values
wmm = WMM(latitude=MUNICH_LATITUDE, longitude=MUNICH_LONGITUDE, height=MUNICH_HEIGHT)
REFERENCE_MAGNETIC_VECTOR = wmm.geodetic_vector

# Gravitational values
NORMAL_GRAVITY = WGS().normal_gravity(MUNICH_LATITUDE, MUNICH_HEIGHT)
REFERENCE_GRAVITY_VECTOR = np.array([0.0, 0.0, NORMAL_GRAVITY])

SAMPLING_FREQUENCY = 100.0

def __gaussian_filter(in_array: np.ndarray, size: int = 10, sigma: float = 1.0) -> np.ndarray:
    """
    Gaussian filter over an array

    This function tries to mimic the behavior of Scipy's function
    ``gaussian_filter1d``. It is re-implemented here, in order to avoid the
    dependency on Scipy.

    Parameters
    ----------
    in_array : np.ndarray
        Input array to be filtered.
    size : int, default: 10
        Size of Kernel used over the input array.
    sigma : float, default: 1.0
        Standard deviation of Gaussian Kernel.

    Returns
    -------
    y : np.ndarray
        Filtered array.
    """
    x = np.linspace(-sigma*4, sigma*4, size)
    phi_x = np.exp(-0.5*x**2/sigma**2)
    phi_x /= phi_x.sum()
    if in_array.ndim < 2:
        return np.correlate(in_array, phi_x, mode='same')
    return np.array([np.correlate(col, phi_x, mode='same') for col in in_array.T]).T

def random_angpos(num_samples: int = 500, max_positions: int = 4, num_axes: int = 3, span: list = None, **kwargs) -> np.ndarray:
    """
    Random angular positions

    Create an array of synthetic random angular positions with respect to each
    axis of a three-dimensional global coordinate frame.

    These angular positions are "simulated" by creating a random number of
    positions per axis, extend them for several samples, and then smoothing
    them with a gaussian filter.

    This creates smooth transitions between the different angular positions.

    .. warning::

        It must be stressed, these are **ANGULAR** positions, not **linear**.
        They are measured in radians, and **do not** represent translations or
        positions in a three-dimensional space.

    Parameters
    ----------
    num_samples : int, default: 500
        Number of samples to generate. Set it to minimum 50, so that the
        gaussian filter can be applied.
    max_positions : int, default: 4
        Maximum number of rotations per axis.
    num_axes : int, default: 3
        Number of axes required.
    span : list or tuple, default: [-pi/2, pi/2]
        Span (minimum to maximum) of the random values.
    rng : np.random.Generator, default: np.random.default_rng(42)
        Random number generator to use. If not given, it uses the default
        random number generator with seed 42.

    Returns
    -------
    angular_positions: np.ndarray
        M-by-3 Array of angular positions.
    """
    rng = kwargs.pop('rng', GENERATOR)
    span = span if isinstance(span, (list, tuple)) else [-0.5*np.pi, 0.5*np.pi]
    all_angs = [rng.uniform(span[0], span[1], rng.integers(1, max_positions)) for _ in np.arange(num_axes)]
    angular_positions = np.zeros((num_samples, num_axes))
    for j, angs in enumerate(all_angs):
        # Create angular positions per axis
        num_angs = len(angs)
        idxs = np.sort(rng.integers(0, num_samples, 2*num_angs)).reshape((num_angs, 2))
        for i, idx in enumerate(idxs):
            # Extend each angular position for several samples
            angular_positions[idx[0]:idx[1], j] = angs[i]
    smoothed_angular_positions = __gaussian_filter(angular_positions, size=kwargs.pop('gauss_size', 50 if num_samples > 50 else num_samples//5), sigma=5)
    return smoothed_angular_positions

class Sensors:
    """
    Generate synthetic sensor data of a hypothetical strapdown inertial
    navigation system.

    It generates data of a 9-DOF IMU (3-axis gyroscopes, 3-axis accelerometers,
    and 3-axis magnetometers) from a given array of orientations as quaternions.

    The accelerometer data is given as m/s^2, the gyroscope data as rad/s, and
    the magnetometer data as nT.

    If no quaternions are provided, it generates random angular positions and
    computes the corresponding quaternions.

    The sensor data can be accessed as attributes of the object. For example,
    the gyroscope data can be accessed as ``sensors.gyroscopes``.

    Simulating N observations, the most used attributes are:

    - ``gyroscopes``: N-by-3 array with gyroscope data, as rad/s.
    - ``accelerometers``: N-by-3 array with accelerometer data, as m/s^2.
    - ``magnetometers``: N-by-3 array with magnetometer data, as nT.
    - ``quaternions``: N-by-4 array with orientations as quaternions.
    - ``rotations``: N-by-3-by-3 array with orientations as 3x3 Rotation
      matrices.
    - ``angular_positions``: N-by-3 array with orientations as Euler angles
      (roll, pitch, yaw.)
    - ``ang_vel``: N-by-3 array with angular velocities around the X-, Y-, and
      Z-axes. Obtained from differentiation of the orientations.
    - ``frequency``: Sampling frequency of the data, in Hz.

    Parameters
    ----------
    quaternions : ahrs.QuaternionArray, default: None
        Array of orientations as quaternions.
    num_samples : int, default: 500
        Number of samples to generate.
    freq : float, default: 100.0
        Sampling frequency, in Hz, of the data.
    in_degrees : bool, default: False
        If True, the gyroscope data is generated in degrees per second.
        Otherwise in radians per second.
    normalized_mag : bool, default: False
        If True, the magnetometer data is normalized to unit norm.
    reference_gravitational_vector : np.ndarray, default: None
        Reference gravitational vector. If None, it uses the default reference
        gravitational vector of ``ahrs.utils.WGS()``.
    reference_magnetic_vector : np.ndarray, default: None
        Reference magnetic vector. If None, it uses the default reference
        magnetic vector of ``ahrs.utils.WMM()``.
    gyr_noise : float
        Standard deviation of the gyroscope noise. If None given, it is
        generated from a normal distribution with zero mean. It is then scaled
        to be in the same units as the gyroscope data.
    acc_noise : float
        Standard deviation of the accelerometer noise. If None given, it is
        generated from a normal distribution with zero mean. It is then scaled
        to be in the same units as the accelerometer data.
    mag_noise : float
        Standard deviation of the magnetometer noise. If None given, it is
        generated from a normal distribution with zero mean. It is then scaled
        to be in the same units as the magnetometer data.
    rng : np.random.Generator, default: np.random.default_rng(42)
        Random number generator to use. If not given, it uses the default
        random number generator with seed 42.

    Examples
    --------
    >>> sensors = Sensors(num_samples=1000)
    >>> sensors.gyroscopes.shape
    (1000, 3)
    >>> sensors.accelerometers.shape
    (1000, 3)
    >>> sensors.magnetometers.shape
    (1000, 3)
    >>> sensors.quaternions.shape
    (1000, 4)

    """
    def __init__(self, quaternions: QuaternionArray = None, num_samples: int = 500, freq: float = SAMPLING_FREQUENCY, **kwargs):
        self.frequency = freq
        self.in_degrees = kwargs.get('in_degrees', False)
        self.normalized_mag = kwargs.get('normalized_mag', False)
        self.rng = kwargs.pop('rng', GENERATOR)

        # Reference earth frames
        self.reference_gravitational_vector = kwargs.get('reference_gravitational_vector', REFERENCE_GRAVITY_VECTOR)
        self.reference_magnetic_vector = kwargs.get('reference_magnetic_vector', REFERENCE_MAGNETIC_VECTOR)

        # Spectral noise density
        self.gyr_noise_default_std_deviation = abs(self.rng.standard_normal(3) * 0.1) * RAD2DEG
        self.acc_noise_default_std_deviation = np.linalg.norm(REFERENCE_GRAVITY_VECTOR) * 0.01
        self.mag_noise_default_std_deviation = np.linalg.norm(REFERENCE_MAGNETIC_VECTOR) * 0.005
        self.gyr_noise = kwargs.get('gyr_noise', self.gyr_noise_default_std_deviation)
        self.acc_noise = kwargs.get('acc_noise', self.acc_noise_default_std_deviation)
        self.mag_noise = kwargs.get('mag_noise', self.mag_noise_default_std_deviation)

        # Orientations as quaternions
        if quaternions is None:
            # Orientations were NOT given
            self.num_samples = num_samples
            # Generate orientations (angular positions)
            self.angular_positions = random_angpos(num_samples=self.num_samples,
                                                   span=kwargs.get("span", (-np.pi, np.pi)),
                                                   max_positions=20,
                                                   rng=self.rng)
            if 'yaw' in kwargs:
                self.angular_positions[:, 2] = kwargs.get('yaw') * DEG2RAD
            self.quaternions = QuaternionArray(rpy=self.angular_positions)
            # Estimate angular velocities
            self.ang_vel = self.angular_velocities(self.angular_positions, self.frequency)
        else:
            # Orientations were given (as quaternions)
            # Define angular positions and velocities
            self.quaternions = QuaternionArray(quaternions)
            self.num_samples = self.quaternions.shape[0]
            self.angular_positions = self.quaternions.to_angles()
            self.ang_vel = np.r_[np.zeros((1, 3)), self.quaternions.angular_velocities(1/self.frequency)]
        # Rotation Matrices
        self.rotations = self.quaternions.to_DCM()

        # Set empty arrays
        self.gyroscopes = None
        self.accelerometers = np.zeros((self.num_samples, 3))
        self.magnetometers = np.zeros((self.num_samples, 3))
        self.magnetometers_nd = np.zeros((self.num_samples, 3))
        self.magnetometers_enu = np.zeros((self.num_samples, 3))

        # Generate MARG data
        self.generate(self.rotations)

    def angular_velocities(self, angular_positions: np.ndarray, frequency: float) -> np.ndarray:
        """Compute angular velocities"""
        Qts = angular_positions if isinstance(angular_positions, QuaternionArray) else QuaternionArray(rpy=angular_positions)
        angvels = Qts.angular_velocities(1/frequency)
        return np.vstack((angvels[0], angvels))

    def generate(self, rotations: np.ndarray) -> None:
        """Compute synthetic data"""
        # Angular velocities measured in the local frame
        self.gyroscopes = np.copy(self.ang_vel) * RAD2DEG

        # Add gyro biases: uniform random constant biases within 1/200th of the full range of the gyroscopes
        self.biases_gyroscopes = (self.rng.random(3)-0.5) * np.ptp(self.gyroscopes)/200
        if not self.in_degrees:
            self.biases_gyroscopes *= DEG2RAD
        self.gyroscopes += self.biases_gyroscopes

        # Accelerometers and magnetometers are measured w.r.t. global frame (inverse of the local frame)
        self.reference_magnetic_vector_nd = np.array([np.cos(wmm.I * DEG2RAD), 0.0, np.sin(wmm.I * DEG2RAD)])
        self.reference_magnetic_vector_enu = ned2enu(self.reference_magnetic_vector)
        for i in np.arange(self.num_samples):
            self.accelerometers[i] = rotations[i].T @ self.reference_gravitational_vector
            self.magnetometers[i] = rotations[i].T @ self.reference_magnetic_vector
            self.magnetometers_nd[i] = rotations[i].T @ self.reference_magnetic_vector_nd
            self.magnetometers_enu[i] = rotations[i].T @ self.reference_magnetic_vector_enu

        # Add noise
        if self.mag_noise < np.ptp(self.magnetometers):
            self.mag_noise = np.linalg.norm(REFERENCE_MAGNETIC_VECTOR) * 0.005
        self.gyroscopes += self.rng.standard_normal((self.num_samples, 3)) * self.gyr_noise
        self.accelerometers += self.rng.standard_normal((self.num_samples, 3)) * self.acc_noise
        self.magnetometers += self.rng.standard_normal((self.num_samples, 3)) * self.mag_noise
        self.magnetometers_nd += self.rng.standard_normal((self.num_samples, 3)) * self.mag_noise
        self.magnetometers_enu += self.rng.standard_normal((self.num_samples, 3)) * self.mag_noise

        if not self.in_degrees:
            self.gyroscopes *= DEG2RAD
            self.biases_gyroscopes *= DEG2RAD
        if self.normalized_mag:
            self.magnetometers /= np.linalg.norm(self.magnetometers, axis=1, keepdims=True)
            self.magnetometers_nd /= np.linalg.norm(self.magnetometers_nd, axis=1, keepdims=True)
            self.magnetometers_enu /= np.linalg.norm(self.magnetometers_enu, axis=1, keepdims=True)



# -*- coding: utf-8 -*-
"""
.. versionadded:: 0.4.0

The Unscented Kaman Filter (UKF) was first proposed by S. Julier and J. Uhlmann
:cite:p:`julier1997` as an alternative to the Kalman Fiter for nonlinear
systems.

The UKF approximates the mean and covariance of the state distribution using a
set of discretely sampled points, called the **Sigma Points**, obtained through
a deterministic sampling technique called the `Unscented Transform
<https://en.wikipedia.org/wiki/Unscented_transform>`_.

Contrary to the EKF, the UKF does not linearize the models, but uses each of
the sigma points as an input to the state transition and measurement functions
to get a new set of transformed state points, thus avoiding the need for
Jacobians, and yielding an accuracy similar to the KF for linear systems.

The UKF offers significant advantages over the EKF in terms of handling
nonlinearities, achieving higher-order accuracy, robustness to initial
estimates, and consistent performance.

However, the UKF has disadvantages related to computational complexity, memory
requirements, and parameter tuning. These factors can make the UKF less
suitable for certain applications, particularly those with limited
computational resources.

The implementation in this module is based on the UKF algorithm for nonlinear
estimations proposed by Wan and van de Merwe :cite:p:`wan2000`, and further
developed by Kraft :cite:p:`kraft2003` and Klingbeil :cite:p:`klingbeil2006`
for orientation estimation using quaternions.

**Kalman Filter**

We have a `discrete system <https://en.wikipedia.org/wiki/Discrete_system>`_,
whose `states <https://en.wikipedia.org/wiki/State_(computer_science)>`_ are
described by a vector :math:`\\mathbf{x}_t` at each time :math:`t`.

This vector has :math:`n` items, which quantify the position, velocity,
orientation, etc. Basically, anything that can be measured or estimated can be
a state, as long as it can be described numerically.

Knowing how the state was at time :math:`t-1`, we want to predict how the state
is at time :math:`t`. In addition, we also have a set of measurements
:math:`\\mathbf{z}_t`, that can be used to improve the prediction of the state.

The traditional `Kalman filter <https://en.wikipedia.org/wiki/Kalman_filter>`_,
as described by :cite:p:`kalman1960` computes a state in two steps:

1. The **prediction step** computes a guess of the current state,
   :math:`\\hat{\\mathbf{x}}_t`, and its covariance :math:`\\hat{\\mathbf{P}}_t`,
   at time :math:`t`, given the previous state and covariance at time :math:`t-1`.

.. math::
    \\begin{array}{rcl}
    \\hat{\\mathbf{x}}_t &=& \\mathbf{F}\\mathbf{x}_{t-1} + \\mathbf{Bu}_t \\\\
    \\hat{\\mathbf{P}}_t &=& \\mathbf{F}\\mathbf{P}_{t-1}\\mathbf{F}^T + \\mathbf{Q}_t
    \\end{array}

2. The **correction step** improves this prediction using a measurement (or set
   of measurements) :math:`\\mathbf{z}_t` at time :math:`t`.

.. math::
    \\begin{array}{rcl}
    \\mathbf{v}_t &=& \\mathbf{z}_t - \\mathbf{H}\\hat{\\mathbf{x}}_t \\\\
    \\mathbf{S}_t &=& \\mathbf{H} \\hat{\\mathbf{P}}_t \\mathbf{H}^T + \\mathbf{R} \\\\
    \\mathbf{K}_t &=& \\hat{\\mathbf{P}}_t \\mathbf{H}^T \\mathbf{S}_t^{-1} \\\\
    \\mathbf{x}_t &=& \\hat{\\mathbf{x}}_t + \\mathbf{K}_t \\mathbf{v}_t \\\\
    \\mathbf{P}_t &=& \\hat{\\mathbf{P}}_t - \\mathbf{K}_t \\mathbf{S}_t \\mathbf{K}_t^T
    \\end{array}

The Kalman filter, however, is limited to `linear systems
<https://en.wikipedia.org/wiki/Linear_system>`_, rendering the process above
inapplicable to `nonlinear systems <https://en.wikipedia.org/wiki/Nonlinear_system>`_
like our attitude estimation problem.

**Extended Kalman Filter**

A common solution to the nonlinearity issue is the `Extended Kalman Filter
<./ekf.html>`_ (EKF), which linearizes the system model and measurement
functions to __approximate__ the terms, allowing the use of the Kalman filter
as if it were a linear system.

In this approach the predicted mean and covariance are computed using the
linearized models:

.. math::

    \\begin{array}{rcl}
    \\hat{\\mathbf{x}}_t &=& \\mathbf{f}(\\mathbf{x}_{t-1}, \\mathbf{u}_t) \\\\
    \\hat{\\mathbf{P}}_t &=& \\mathbf{F}(\\mathbf{x}_{t-1}, \\mathbf{u}_t)\\mathbf{P}_{t-1}\\mathbf{F}^T(\\mathbf{x}_{t-1}, \\mathbf{u}_t) + \\mathbf{Q}_t
    \\end{array}

where :math:`\\mathbf{f}(\\mathbf{x}_{t-1}, \\mathbf{u}_t)` is the nonlinear
dynamic model function, whose Jacobian is:

.. math::

    \\mathbf{F}(\\mathbf{x}_{t-1}, \\mathbf{u}_t) = \\frac{\\partial \\mathbf{f}(\\mathbf{x}_{t-1}, \\mathbf{u}_t)}{\\partial \\mathbf{x}}

whereas the measurement model [#]_ is linearized as:

.. math::

    \\begin{array}{rcl}
    \\mathbf{v}_t &=& \\mathbf{z}_t - \\mathbf{h}(\\mathbf{x}_t) \\\\
    \\mathbf{S}_t &=& \\mathbf{H}(\\mathbf{x}_t) \\hat{\\mathbf{P}}_t \\mathbf{H}^T(\\mathbf{x}_t) + \\mathbf{R}_t \\\\
    \\mathbf{K}_t &=& \\hat{\\mathbf{P}}_t \\mathbf{H}^T(\\mathbf{x}_t) \\mathbf{S}_t^{-1} \\\\
    \\mathbf{x}_t &=& \\hat{\\mathbf{x}}_t + \\mathbf{K}_t \\mathbf{v}_t \\\\
    \\mathbf{P}_t &=& \\big(\\mathbf{I}_4 - \\mathbf{K}_t\\mathbf{H}(\\mathbf{x}_t)\\big)\\hat{\\mathbf{P}}_t
    \\end{array}

where :math:`\\mathbf{h}(\\mathbf{x}_t)` is the nonlinear measurement model
function, whose Jacobian is:

.. math::

    \\mathbf{H}(\\hat{\\mathbf{x}}_t) = \\frac{\\partial \\mathbf{h}(\\mathbf{x}_t)}{\\partial \\mathbf{x}}

Unfortunately, these approximations can introduce large errors in the posterior
mean and covariance of the transformed random variable, which may lead to
sub-optimal performance.

To avoid these issues, a solution using unscented transforms was proposed by
Julier and Uhlmann :cite:p:`julier1997`, which is the basis for the Unscented
Kalman Filter (UKF).

Unscented Kalman Filter
------------------------

**Unscented Transform**

The UKF is a type of Kalman filter that replaces the linearization with a
deterministic sampling technique called the `Unscented Transform
<https://en.wikipedia.org/wiki/Unscented_transform>`_.

This transformation generates a set of points that capture the mean and
covariance of the state distribution, called the **Sigma Points**.

Each of the sigma points is used as an input to the state transition and
measurement functions to get a new set of transformed state points.

.. epigraph::

   The unscented transformation ... is founded on the intuition that it is
   easier to approximate a Gaussian distribution than it is to approximate an
   arbitrary nonlinear function or transformation.

   -- Jeffrey K. Uhlmann

Imagine there is a set of random points :math:`\\mathbf{x}` with mean
:math:`\\bar{\\mathbf{x}}`, and covariance :math:`\\mathbf{P_{xx}}`, and there
is another set of random points :math:`\\mathbf{y}` related to
:math:`\\mathbf{x}` by a nonlinear function :math:`\\mathbf{y} = f(\\mathbf{x})`.

Our goal is to find the mean :math:`\\bar{\\mathbf{y}}` and covariance
:math:`\\mathbf{P_{yy}}` of :math:`\\mathbf{y}`. The unscented transform
approximates them by sampling a set of points from :math:`\\mathbf{x}` and
applying the nonlinear function :math:`f` to each of the sampled points.

Information about the distribution can be captured using a small number of
points :cite:p:`julier1997`. The samples are not drawn at random but according
to a deterministic method.

**Sigma Points**

The :math:`n`-dimensional random variable :math:`\\mathbf{x}` with mean
:math:`\\bar{\\mathbf{x}}` and covariance :math:`\\mathbf{P_{xx}}` is
approximated by :math:`2n + 1` points computed with:

.. math::

    \\begin{array}{rcl}
    \\mathcal{X}_0 &=& \\bar{\\mathbf{x}} \\\\
    \\mathcal{X}_i &=& \\bar{\\mathbf{x}} + \\Big(\\sqrt{(n + \\lambda)\\mathbf{P_{xx}}}\\Big)_i\\\\
    \\mathcal{X}_{i+n} &=& \\bar{\\mathbf{x}} - \\Big(\\sqrt{(n + \\lambda)\\mathbf{P_{xx}}}\\Big)_i
    \\end{array}

where :math:`(\\sqrt{(n + \\lambda)\\mathbf{P_{xx}}})_i` is the :math:`i`-th
column of the `matrix square root <https://en.wikipedia.org/wiki/Square_root_of_a_matrix>`_,
and :math:`\\lambda=\\alpha^2(n + \\kappa) - n` is a scaling parameter.

:math:`\\alpha` determines the spread of the sigma points around the mean,
usually set to :math:`0.001`, and :math:`\\kappa` is a secondary scaling
parameter, usually set to :math:`0` :cite:p:`wan2000`.

But, how do we obtain the matrix form of the `square root
<https://en.wikipedia.org/wiki/Square_root_of_a_matrix>`_ of
:math:`(n + \\lambda)\\mathbf{P_{xx}}`?

Because :math:`\\mathbf{P_{xx}}` is a covariance matrix, it means it is
symmetric and positive-definite.

The `Cholesky decomposition <https://en.wikipedia.org/wiki/Cholesky_decomposition>`_
of a `positive-definite matrix <https://en.wikipedia.org/wiki/Definite_matrix>`_
:math:`\\mathbf{A}` is a lower triangular matrix :math:`\\mathbf{L}` such that
:math:`\\mathbf{A} = \\mathbf{LL}^T`.

The square root of :math:`(n + \\lambda)\\mathbf{P_{xx}}` is then
:math:`\\mathbf{L}=\\mathrm{chol}((n + \\lambda)\\mathbf{P_{xx}})`.

The Cholesky decomposition is preferred because:

- It efficiently computes (roughly :math:`\\frac{n^3}{3}` operations for an
  :math:`n\\times n` matrix) the lower triangular matrix :math:`\\mathbf{L}`.
- It's numerically stable.
- It naturally handles the positive-definiteness requirement of covariance
  matrices.

Therefore, we first calculate the Cholesky decomposition of
:math:`(n + \\lambda)\\mathbf{P_{xx}}`, and then we obtain the Sigma Points by
adding and subtracting the columns of :math:`\\mathbf{L}` to the mean.

.. math::
   :label: sigma_points

    \\begin{array}{rcl}
    \\mathcal{X}_0 &=& \\bar{\\mathbf{x}} \\\\
    \\mathcal{X}_i &=& \\bar{\\mathbf{x}} + \\mathbf{L}_i \\\\
    \\mathcal{X}_{i+n} &=& \\bar{\\mathbf{x}} - \\mathbf{L}_i
    \\end{array}

where :math:`\\mathbf{L}` is the Cholesky decomposition of
:math:`(n + \\lambda)\\mathbf{P_{xx}}`, as mentioned above.

We pass these sigma points through the nonlinear function :math:`f` to get the
transformed points :math:`\\mathcal{Y}`.

.. math::
   :label: ukf_process_model

    \\mathcal{Y} = f(\\mathcal{X})

Their **mean** is given by their wieghted sum:

.. math::
   :label: ukf_predicted_state_mean

    \\bar{\\mathbf{y}} = \\sum_{i=0}^{2n} w_i^{(m)} \\mathcal{Y}_i

And their **covariance** by their weighted outer product:

.. math::
   :label: ukf_predicted_state_covariance

    \\mathbf{P_{yy}} = \\sum_{i=0}^{2n} w_i^{(c)} (\\mathcal{Y}_i - \\bar{\\mathbf{y}})(\\mathcal{Y}_i - \\bar{\\mathbf{y}})^T + \\mathbf{Q}

with :math:`\\mathbf{Q}` being the :math:`n\\times n` process noise covariance
matrix.

The weights :math:`\\mathbf{w}` are computed as:

.. math::

    \\begin{array}{rcl}
    w_0^{(m)} &=& \\frac{\\lambda}{n + \\lambda} \\\\
    w_0^{(c)} &=& \\frac{\\lambda}{n + \\lambda} + (1 - \\alpha^2 + \\beta) \\\\
    w_i^{(m)} = w_i^{(c)} &=& \\frac{1}{2(n + \\lambda)} \\quad \\text{for} \\quad i=1,2,\\ldots,2n
    \\end{array}

The weights :math:`\\mathbf{w}^{(m)}` are used to compute the mean, and the
weights :math:`\\mathbf{w}^{(c)}` are used to compute the covariance.

The constant :math:`\\beta` is used to incorporate prior knowledge about the
distribution of the random variable, and is usually set to :math:`2` for
Gaussian distributions :cite:p:`wan2000`.

**UKF Summary**

Given the previous state :math:`\\mathbf{x}_{t-1}`, its covariance matrix
:math:`\\mathbf{P}_{t-1}`, a control vector :math:`\\mathbf{u}_t`, and a vector
with the most recent :math:`m` measurements :math:`\\mathbf{z}_t`, the UKF
algorithm can be summarized as follows:

**Prediction**:

1. Compute Sigma Points.

.. math::

    \\mathcal{X}(\\mathbf{x}_{t-1}, \\mathbf{P}_{t-1}) = \\Big\\{ \\mathcal{X}_0 \\; , \\quad\\mathcal{X}_i \\; , \\quad\\mathcal{X}_{i+n} \\Big\\}

2. Propagate Sigma Points through Process model to get Transformed Sigma Points.

.. math::

    \\mathcal{Y} = f(\\mathcal{X})

3. Predict State Mean and Covariance.

.. math::

    \\begin{array}{rcl}
    \\bar{\\mathbf{y}} &=& \\sum_{i=0}^{2n} w_i^{(m)} \\mathcal{Y}_i \\\\ \\\\
    \\mathbf{P_{yy}} &=& \\sum_{i=0}^{2n} w_i^{(c)} (\\mathcal{Y}_i - \\bar{\\mathbf{y}})(\\mathcal{Y}_i - \\bar{\\mathbf{y}})^T + \\mathbf{Q}
    \\end{array}

**Correction**:

4. Instantiate each Transformed Sigma Point through Measurement model.

.. math::

    \\mathcal{Z} = h(\\mathcal{Y})

5. Compute Predicted Measurement Mean and Covariance.

.. math::

    \\begin{array}{rcl}
    \\bar{\\mathbf{z}} &=& \\sum_{i=0}^{2n} w_i^{(m)} \\mathcal{Z}_i \\\\ \\\\
    \\mathbf{P_{zz}} &=& \\sum_{i=0}^{2n} w_i^{(c)} (\\mathcal{Z}_i - \\bar{\\mathbf{z}})(\\mathcal{Z}_i - \\bar{\\mathbf{z}})^T + \\mathbf{R}
    \\end{array}

6. Compute Cross-Covariance.

.. math::

    \\mathbf{P_{yz}} = \\sum_{i=0}^{2n} w_i^{(c)} (\\mathcal{Y}_i - \\bar{\\mathbf{y}})(\\mathcal{Z}_i - \\bar{\\mathbf{z}})^T

7. Compute Kalman gain.

.. math::

    \\mathbf{K} = \\mathbf{P_{yz}} \\mathbf{P_{zz}}^{-1}

8. Compute Innovation (residual.)

.. math::

    \\mathbf{v}_t = \\mathbf{z}_t - \\bar{\\mathbf{z}}

9. Update (Correct) State and Covariance

.. math::

    \\begin{array}{rcl}
    \\mathbf{x}_t &=& \\bar{\\mathbf{y}} + \\mathbf{K} \\mathbf{v}_t \\\\ \\\\
    \\mathbf{P}_t &=& \\mathbf{P_{yy}} - \\mathbf{K} \\mathbf{P_{zz}} \\mathbf{K}^T
    \\end{array}

UKF for Attitude Estimation
---------------------------

In this implementation, we build the simplest UKF for attitude estimation using
gyroscopes and accelerometers (and magnetometer, if available), so that we can
focus on the details of the algorithm. Once the basic structure is understood,
we could extend the model to create more complex systems.

**PREDICTION MODEL**

We start by defining the vectors of the Prediction Model, a.k.a. **Process
Model**:

.. math::

    \\begin{array}{rcl}
    \\mathbf{x} &=& \\begin{bmatrix} q_w & q_x & q_y & q_z \\end{bmatrix}^T \\\\ \\\\
    \\mathbf{u} &=& \\begin{bmatrix} \\omega_x & \\omega_y & \\omega_z \\end{bmatrix}^T
    \\end{array}

The state vector :math:`\\mathbf{x}_t\\in\\mathbb{R}^4` has the elements of a
quaternion representing the orientation at any time :math:`t`.

We have added the **control vector** :math:`\\mathbf{u}_t\\in\\mathbb{R}^3`
containing the angular velocity readings from a tri-axial gyroscope. This
control vector is used to propagate the state vector through the process model.
Normally, it is ignored when the gyroscope is not available or when we assume
the model is not affected by external forces, but this is not the case in our
implementation.

The quaternion describing the orientation is also known as a `versor
<https://en.wikipedia.org/wiki/Versor>`_, and it is a unit quaternion, meaning
:math:`\\|\\mathbf{q}\\|=1`. Thus, we normalize the quaternion after each
transformation.

.. note::

    Notice we don't extend the state vector to include the gyroscope biases
    like other models do. For the sake of simplicity we don't estimate these
    biases, and assume the sensor readings are already calibrated.

**Sigma Points**

Given the previous state and covariance, the sigma points are computed first.

Using the cholesky decomposition we obtain the **matrix square root**:

.. math::

    \\mathbf{L} = \\mathrm{chol}\\Big(\\sqrt{(n + \\lambda)\\mathbf{P_{xx}}}\\Big)

where the previous covariance is set as :math:`\\mathbf{P_{xx}}=\\mathbf{P}_{t-1}`,
:math:`n=4` is the number of items in the state vector :math:`\\mathbf{x}`, and
:math:`\\lambda=\\alpha^2(n + \\kappa) - n` is the scaling parameter.

.. hint::

    Using the default values :math:`\\alpha=0.001`, and :math:`\\kappa=0`, we
    get:

    .. math::

        \\lambda = 0.001^2 (4 + 0) - 4 = -3.999996

    which yields :math:`\\mathbf{L} = \\mathrm{chol}\\big(\\sqrt{0.000004\\mathbf{P}_{t-1}}\\big)`.

Then, we compute the sigma points using the equations:

.. math::

    \\begin{array}{rcl}
    \\mathcal{X}_0 &=& \\mathbf{x}_{t-1} \\\\
    \\mathcal{X}_i &=& \\mathbf{x}_{t-1} + \\mathbf{L}_i \\\\
    \\mathcal{X}_{i+n} &=& \\mathbf{x}_{t-1} - \\mathbf{L}_i
    \\end{array}

The first sigma point :math:`\\mathcal{X}_0` is always equal to the previous
state :math:`\\mathbf{x}_{t-1}`. The rest are obtained by adding and
subtracting the columns of :math:`\\mathbf{L}` to the mean.

Because the state vector has 4 items, we obtain a set of 9 sigma points:

.. math::

    \\begin{array}{rcl}
    \\mathcal{X} &=&
    \\begin{Bmatrix}
        \\big| & \\big| & \\big| & \\big| & \\big| & \\big| & \\big| & \\big| & \\big| \\\\
        \\mathcal{X}_0 & \\mathcal{X}_1 & \\mathcal{X}_2 & \\mathcal{X}_3 & \\mathcal{X}_4 & \\mathcal{X}_5 & \\mathcal{X}_6 & \\mathcal{X}_7 & \\mathcal{X}_8 \\\\
        \\big| & \\big| & \\big| & \\big| & \\big| & \\big| & \\big| & \\big| & \\big|
    \\end{Bmatrix} \\\\ \\\\
    &=&
    \\begin{Bmatrix}
        q_w & q_w + \\mathbf{L}_{1,1} & q_w + \\mathbf{L}_{1,2} & q_w + \\mathbf{L}_{1,3} & q_w + \\mathbf{L}_{1,4} & q_w - \\mathbf{L}_{1,1} & q_w - \\mathbf{L}_{1,2} & q_w - \\mathbf{L}_{1,3} & q_w - \\mathbf{L}_{1,4} \\\\
        q_x & q_x + \\mathbf{L}_{2,1} & q_x + \\mathbf{L}_{2,2} & q_x + \\mathbf{L}_{2,3} & q_x + \\mathbf{L}_{2,4} & q_x - \\mathbf{L}_{2,1} & q_x - \\mathbf{L}_{2,2} & q_x - \\mathbf{L}_{2,3} & q_x - \\mathbf{L}_{2,4} \\\\
        q_y & q_y + \\mathbf{L}_{3,1} & q_y + \\mathbf{L}_{3,2} & q_y + \\mathbf{L}_{3,3} & q_y + \\mathbf{L}_{3,4} & q_y - \\mathbf{L}_{3,1} & q_y - \\mathbf{L}_{3,2} & q_y - \\mathbf{L}_{3,3} & q_y - \\mathbf{L}_{3,4} \\\\
        q_z & q_z + \\mathbf{L}_{4,1} & q_z + \\mathbf{L}_{4,2} & q_z + \\mathbf{L}_{4,3} & q_z + \\mathbf{L}_{4,4} & q_z - \\mathbf{L}_{4,1} & q_z - \\mathbf{L}_{4,2} & q_z - \\mathbf{L}_{4,3} & q_z - \\mathbf{L}_{4,4}
    \\end{Bmatrix}
    \\end{array}

The estimation process is done as a two-step filter consisting of an `attitude
propagation <./angular.html>`_ (using the gyroscope) and a correction (using
the accelerometer.)

**Attitude Propagation**

Based on the time spent between :math:`t-1` and :math:`t` (known as
the time step :math:`\\Delta t`) we can compute the angular displacement
:math:`\\boldsymbol\\Delta_\\theta` using the measured angular velocities
:math:`\\boldsymbol\\omega` from the gyroscopes, and add it to the previous
attitude :math:`\\mathbf{x}_{t-1}` to obtain the predicted attitude
:math:`\\hat{\\mathbf{x}}_t`:

.. math::

    \\begin{array}{rcl}
    \\hat{\\mathbf{x}}_t &=& \\mathbf{x}_{t-1} + \\boldsymbol\\Delta_\\theta \\\\
    &=& \\mathbf{x}_{t-1} + \\int_{t-1}^t\\boldsymbol\\omega\\, dt
    \\end{array}

This is called **attitude propagation**. However, this is a nonlinear operation
and we cannot use it in the Kalman Filter. Therefore, we approximate it to
define our required linear **Process Model**:

.. math::

    \\begin{array}{rcl}
    \\hat{\\mathbf{x}}_t &=& f(\\mathbf{x}_{t-1}, \\boldsymbol\\omega_t) \\\\
    &=&\\Big[\\mathbf{I}_4 + \\frac{\\Delta t}{2}\\boldsymbol\\Omega_t(\\boldsymbol\\omega_t)\\Big]\\mathbf{x}_{t-1} \\\\
    &=&
    \\begin{bmatrix}
    1 & -\\frac{\\Delta t}{2}\\omega_x & -\\frac{\\Delta t}{2}\\omega_y & -\\frac{\\Delta t}{2}\\omega_z \\\\
    \\frac{\\Delta t}{2}\\omega_x & 1 & \\frac{\\Delta t}{2}\\omega_z & -\\frac{\\Delta t}{2}\\omega_y \\\\
    \\frac{\\Delta t}{2}\\omega_y & -\\frac{\\Delta t}{2}\\omega_z & 1 & \\frac{\\Delta t}{2}\\omega_x \\\\
    \\frac{\\Delta t}{2}\\omega_z & \\frac{\\Delta t}{2}\\omega_y & -\\frac{\\Delta t}{2}\\omega_x & 1
    \\end{bmatrix}
    \\begin{bmatrix}
        q_w \\\\ q_x \\\\ q_y \\\\ q_z
    \\end{bmatrix} \\\\
    \\begin{bmatrix}\\hat{q_w} \\\\ \\hat{q_x} \\\\ \\hat{q_y} \\\\ \\hat{q_z}\\end{bmatrix}
    &=&
    \\begin{bmatrix}
        q_w - \\frac{\\Delta t}{2} \\omega_x q_x - \\frac{\\Delta t}{2} \\omega_y q_y - \\frac{\\Delta t}{2} \\omega_z q_z\\\\
        q_x + \\frac{\\Delta t}{2} \\omega_x q_w - \\frac{\\Delta t}{2} \\omega_y q_z + \\frac{\\Delta t}{2} \\omega_z q_y\\\\
        q_y + \\frac{\\Delta t}{2} \\omega_x q_z + \\frac{\\Delta t}{2} \\omega_y q_w - \\frac{\\Delta t}{2} \\omega_z q_x\\\\
        q_z - \\frac{\\Delta t}{2} \\omega_x q_y + \\frac{\\Delta t}{2} \\omega_y q_x + \\frac{\\Delta t}{2} \\omega_z q_w
    \\end{bmatrix}
    \\end{array}

where the rotation operator :math:`\\big[\\mathbf{I}_4 + \\frac{\\Delta t}{2}
\\boldsymbol\\Omega_t(\\boldsymbol\\omega_t)\\big]` is a truncation up to the
second term of the Taylor series expansion of
:math:`\\int_{t-1}^t\\boldsymbol\\omega\\, dt`.

.. note::

    For more details about this linear operation, please refer to the
    documentation of the `Attitude from Angular Rate <./angular.html>`_.

We assume in this description, that the time step :math:`\\Delta t` is constant.
However, it can be changed at any time during the implementation.

We propagate each of the sigma points through the process model
:math:`f` to get a new set of transformed state points :math:`\\mathcal{Y}`:

.. math::

    \\mathcal{Y} =
    \\begin{Bmatrix}
        \\big| & \\big| & & \\big| \\\\
        f(\\mathcal{X}_0, \\boldsymbol\\omega_t) &
        f(\\mathcal{X}_1, \\boldsymbol\\omega_t) &
        \\cdots &
        f(\\mathcal{X}_8, \\boldsymbol\\omega_t) \\\\
        \\big| & \\big| & & \\big|
    \\end{Bmatrix}

Every :math:`\\mathcal{Y}_i` describes a quaternion. They must be normalized after the transformation, so that :math:`\\forall i \\in
\\{0, \\ldots, 2n\\} \\;, \\|\\mathcal{Y}_i\\|=1`.

Now we compute the **Predicted State Mean**.

.. math::

    \\boxed{\\bar{\\mathbf{y}} = \\sum_{i=0}^{2n} w_i^{(m)} \\mathcal{Y}_i}

This predicted state represents the mean of the transformed state points as a
quaternion. Therefore, we must normalize it:

.. math::

    \\bar{\\mathbf{y}} \\leftarrow \\frac{\\bar{\\mathbf{y}}}{\\|\\bar{\\mathbf{y}}\\|}

We proceed to compute the **Predicted State Covariance**.

.. math::

    \\boxed{\\mathbf{P_{yy}} = \\sum_{i=0}^{2n} w_i^{(c)} (\\mathcal{Y}_i - \\bar{\\mathbf{y}})(\\mathcal{Y}_i - \\bar{\\mathbf{y}})^T + \\mathbf{Q}}

Notice the product :math:`(\\mathcal{Y}_i - \\bar{\\mathbf{y}})(\\mathcal{Y}_i
- \\bar{\\mathbf{y}})^T` is the `outer product
<https://en.wikipedia.org/wiki/Outer_product>`_ of the vector
:math:`(\\mathcal{Y}_i - \\bar{\\mathbf{y}})`, which results in a
:math:`4\\times 4` matrix.

**MEASUREMENT MODEL**

Our strategy is to use Earth's known physical reference values, rotate them
around the set of predicted orientations :math:`\\mathcal{Y}`, and compare
these against actual sensor readings :math:`\\mathbf{z}`.

Their difference tells us how "off" the predicted orientation is. To ease the
comparison both the references and the sensor readings are `normalized
<https://en.wikipedia.org/wiki/Unit_vector>`_.

The two main physical references are commonly used in attitude estimation, and
we will use them in our **Measurement Model**:

- Earth's gravitational vector :math:`\\mathbf{g}`.
- Earth's geomagnetic vector :math:`\\mathbf{r}`.

In order to perform the rotations, we use the `direction cosine matrix
<../dcm.html>`_, a.k.a. rotation matrix, built **from each predicted
orientation** (transformed points) to rotate the reference vectors to the
sensor frame.

For `Earth's gravitational vector <https://en.wikipedia.org/wiki/Gravity_of_Earth#Direction>`_,
we set the normalized reference on `local tangent plane coordinates
<https://en.wikipedia.org/wiki/Local_tangent_plane_coordinates>`_ (LTP) as:

.. math::

    \\mathbf{g} = \\begin{bmatrix} 0 \\\\ 0 \\\\ 1 \\end{bmatrix}

We assume Earth's gravitational vector is aligned with the Z-axis of the LTP
and, therefore, the X- and Y-axes are equal to zero.

`Earth's Magnetic Field <https://en.wikipedia.org/wiki/Earth%27s_magnetic_field>`_
is not aligned with a particular axis of `Earth's ellipsoid
<https://en.wikipedia.org/wiki/Earth_ellipsoid>`_, and it is fully described
with:

.. math::

    \\mathbf{r} = \\begin{bmatrix} r_x \\\\ r_y \\\\ r_z \\end{bmatrix}

where :math:`r_x`, :math:`r_y`, and :math:`r_z` are the components of the
reference geomagnetic vector, which can be obtained from the `World Magnetic
Model <https://en.wikipedia.org/wiki/World_Magnetic_Model>`_ (WMM)
based on the sensor's geographical location.

This vector represents the direction and intensity of the Earth's magnetic
field at the sensor's location.

The reference geomagnetic vector :math:`\\mathbf{r}` is also normalized to
unit length, so that we can use it as a direction vector too:

.. math::

    \\mathbf{r} = \\frac{\\mathbf{r}}{\\|\\mathbf{r}\\|} = \\begin{bmatrix} r_x \\\\ r_y \\\\ r_z \\end{bmatrix}

**Sensor Readings**

If only a tri-axial accelerometer is available to correct (update) the
predicted attitude, the measurement vector :math:`\\mathbf{z}_t` has normalized
accelerometer readings only.

.. math::

    \\mathbf{z} = \\mathbf{a} = \\begin{bmatrix} a_x \\\\ a_y \\\\ a_z \\end{bmatrix}

However, if both tri-axial accelerometer and magnetometer are available, the
measurement vector :math:`\\mathbf{z}_t` includes both:

.. math::

    \\mathbf{z} = \\begin{bmatrix} a_x \\\\ a_y \\\\ a_z \\\\ m_x \\\\ m_y \\\\ m_z \\end{bmatrix}

Given one predicted orientation :math:`\\mathcal{Y}_i` we rotate a reference
vector :math:`\\mathbf{g}` or :math:`\\mathbf{r}` to get its corresponding
expected sensor reading. We call this the **Measurement Model Function**.

If only the accelerometer is available, the measurement model function yields:

.. math::

    \\begin{array}{rcl}
    h(\\hat{\\mathbf{x}}) &=& \\mathbf{R}(\\hat{\\mathbf{x}})^T \\mathbf{g} \\\\
    &=& \\begin{bmatrix}
        1 - 2(\\hat{q_y}^2 + \\hat{q_z}^2) & 2(\\hat{q_x} \\hat{q_y} + \\hat{q_w} \\hat{q_z}) & 2(\\hat{q_x} \\hat{q_z} - \\hat{q_w} \\hat{q_y}) \\\\
        2(\\hat{q_x} \\hat{q_y} - \\hat{q_w} \\hat{q_z}) & 1 - 2(\\hat{q_x}^2 + \\hat{q_z}^2) & 2(\\hat{q_y} \\hat{q_z} + \\hat{q_w} \\hat{q_x}) \\\\
        2(\\hat{q_x} \\hat{q_z} + \\hat{q_w} \\hat{q_y}) & 2(\\hat{q_y} \\hat{q_z} - \\hat{q_w} \\hat{q_x}) & 1 - 2(\\hat{q_x}^2 + \\hat{q_y}^2)
    \\end{bmatrix}
    \\begin{bmatrix} 0 \\\\ 0 \\\\ 1 \\end{bmatrix} \\\\
    &=& \\begin{bmatrix}
        2(\\hat{q_x} \\hat{q_z} - \\hat{q_w} \\hat{q_y}) \\\\
        2(\\hat{q_y} \\hat{q_z} + \\hat{q_w} \\hat{q_x}) \\\\
        1 - 2(\\hat{q_x}^2 + \\hat{q_y}^2)
    \\end{bmatrix}
    \\end{array}

If both accelerometer and magnetometer are available, it is stacked as:

.. math::

    \\begin{array}{rcl}
    h(\\hat{\\mathbf{x}}) &=& \\begin{bmatrix}
        \\mathbf{R}(\\hat{\\mathbf{x}})^T \\mathbf{g} \\\\
        \\mathbf{R}(\\hat{\\mathbf{x}})^T \\ \\mathbf{r}
    \\end{bmatrix} \\\\
    &=& \\begin{bmatrix}
        2(\\hat{q_x} \\hat{q_z} - \\hat{q_w} \\hat{q_y}) \\\\
        2(\\hat{q_y} \\hat{q_z} + \\hat{q_w} \\hat{q_x}) \\\\
        1 - 2(\\hat{q_x}^2 + \\hat{q_y}^2) \\\\
        r_x\\big(1 - 2(\\hat{q_y}^2 + \\hat{q_z}^2)\\big) + 2r_y(\\hat{q_x} \\hat{q_y} + \\hat{q_w} \\hat{q_z}) + 2r_z(\\hat{q_x} \\hat{q_z} - \\hat{q_w} \\hat{q_y}) \\\\
        2r_x(\\hat{q_x} \\hat{q_y} - \\hat{q_w} \\hat{q_z}) + r_y\\big(1 - 2(\\hat{q_x}^2 + \\hat{q_z}^2)\\big) + 2r_z(\\hat{q_y} \\hat{q_z} + \\hat{q_w} \\hat{q_x}) \\\\
        2r_x(\\hat{q_x} \\hat{q_z} + \\hat{q_w} \\hat{q_y}) + 2r_y(\\hat{q_y} \\hat{q_z} - \\hat{q_w} \\hat{q_x}) + r_z\\big(1 - 2(\\hat{q_x}^2 + \\hat{q_y}^2)\\big)
    \\end{bmatrix}
    \\end{array}

.. note::

    Notice we use the transpose of the rotation matrix to rotate the gravity
    vector from the global frame to the sensor frame (the opposite of what it
    describes.) We do this, so that we can compare it against the accelerometer
    readings in sensor frame.

We execute the measurement model function over each of the predicted sigma
points :math:`\\mathcal{Y}` to get the expected accelerometer readings
:math:`\\mathcal{Z}`:

.. math::

    \\mathcal{Z} =
    \\begin{Bmatrix}
        \\big| & \\big| & & \\big| \\\\
        h(\\mathcal{Y}_0) &
        h(\\mathcal{Y}_1) &
        \\cdots &
        h(\\mathcal{Y}_8) \\\\
        \\big| & \\big| & & \\big|
    \\end{Bmatrix}

With this set of expected accelerometer readings :math:`\\mathcal{Z}` we can
compute the **Measurement Mean**:

.. math::

    \\boxed{\\bar{\\mathbf{z}} = \\sum_{i=0}^{2n} w_i^{(m)} \\mathcal{Z}_i}

And the **Measurement Covariance**:

.. math::

    \\boxed{\\mathbf{P_{zz}} = \\sum_{i=0}^{2n} w_i^{(c)} (\\mathcal{Z}_i - \\bar{\\mathbf{z}})(\\mathcal{Z}_i - \\bar{\\mathbf{z}})^T + \\mathbf{R}}

The **Measurement Noise Covariance Matrix** :math:`\\mathbf{R}` is built based
on whether we have accelerometer readings only, or both accelerometer and
magnetometer readings.

.. math::

    \\mathbf{R} &=&
    \\left\\{
    \\begin{array}{ll}
        \\begin{bmatrix}
            \\sigma_{a_x}^2 & 0 & 0 \\\\
            0 & \\sigma_{a_y}^2 & 0 \\\\
            0 & 0 & \\sigma_{a_z}^2
        \\end{bmatrix}
        & \\mathrm{if}\\; \\mathrm{Acc} \\\\ \\\\
        \\begin{bmatrix}
            \\sigma_{a_x}^2 & 0 & 0 & 0 & 0 & 0 \\\\
            0 & \\sigma_{a_y}^2 & 0 & 0 & 0 & 0 \\\\
            0 & 0 & \\sigma_{a_z}^2 & 0 & 0 & 0 \\\\
            0 & 0 & 0 & \\sigma_{m_x}^2 & 0 & 0 \\\\
            0 & 0 & 0 & 0 & \\sigma_{m_y}^2 & 0 \\\\
            0 & 0 & 0 & 0 & 0 & \\sigma_{m_z}^2
        \\end{bmatrix}
        & \\mathrm{if}\\; \\mathrm{Acc+Mag}
    \\end{array}
    \\right.\\\\

where :math:`\\sigma_{a_x}`, :math:`\\sigma_{a_y}`, and :math:`\\sigma_{a_z}`
are the `standard deviations <https://en.wikipedia.org/wiki/Standard_deviation>`_
of the accelerometer readings, and :math:`\\sigma_{m_x}`, :math:`\\sigma_{m_y}`,
and :math:`\\sigma_{m_z}` are the standard deviations of the magnetometer
readings.

The **Cross-Covariance Matrix** :math:`\\mathbf{P_{yz}}` represents how changes
in the state variables correlate with changes in the measurement variables.
Specifically, it quantifies how errors in the predicted states are related to
errors in the expected measurements.

.. math::

    \\boxed{\\mathbf{P_{yz}} = \\sum_{i=0}^{2n} w_i^{(c)} (\\mathcal{Y}_i - \\bar{\\mathbf{y}})(\\mathcal{Z}_i - \\bar{\\mathbf{z}})^T}

With these matrices we compute the **Kalman Gain**.

.. math::

    \\boxed{\\mathbf{K} = \\mathbf{P_{yz}} \\mathbf{P_{zz}}^{-1}}

Notice this is a much simpler operation than the Extended Kalman Filter (EKF)
where we need to compute the Jacobian matrix.

We compare the expected measurement mean :math:`\\bar{\\mathbf{z}}` against
the actual measurement reading :math:`\\mathbf{z}` to get the **innovation** at
the curent time step :math:`t`:

.. math::

    \\boxed{\\mathbf{v}_t = \\mathbf{z}_t - \\bar{\\mathbf{z}}_t}

This tells us, basically, how "far off" the predicted state is from the actual
measurements.

Finally, we use all this information to correct the state and covariance:

.. math::

    \\boxed{
    \\begin{array}{rcl}
    \\mathbf{x}_t &=& \\bar{\\mathbf{y}} + \\mathbf{K} \\mathbf{v}_t \\\\ \\\\
    \\mathbf{P}_t &=& \\mathbf{P_{yy}} - \\mathbf{K} \\mathbf{P_{zz}} \\mathbf{K}^T
    \\end{array}}

Footnotes
---------
.. [#] The Measurement Model is sometimes called **Observation Model**.

.. seealso::

   - `EKF <./ekf.html>`_ - Extended Kalman Filter for orientation estimation.
   - `AngularRate <./angular.html>`_ - Attitude propagation using angular rate.

"""

import numpy as np
from ..common.quaternion import Quaternion
from ..common.orientation import ecompass
from ..common.orientation import acc2q
from ..utils.core import _assert_numerical_iterable
# Local magnetic reference of Munich, Germany
from ..common.constants import MUNICH_LATITUDE, MUNICH_LONGITUDE, MUNICH_HEIGHT
from ..utils.wmm import WMM

class UKF:
    """
    Unscented Kalman Filter to estimate orientation as Quaternion.

    Examples
    --------
    >>> import numpy as np
    >>> from ahrs.filters import UKF
    >>> from ahrs.common.orientation import ecompass
    >>> ukf = UKF()
    >>> num_samples = 1000              # Assuming sensors have 1000 samples each
    >>> Q = np.zeros((num_samples, 4))  # Allocate array for quaternions
    >>> Q[0] = ecompass(acc_data[0], mag_data[0])   # First sample of tri-axial accelerometer and magnetometer
    >>> for t in range(1, num_samples):
    ...     Q[t] = ukf.update(Q[t-1], gyr_data[t], acc_data[t], mag_data[t])

    The estimation can be simplified by giving all sensor values at the
    construction of the UKF object.

    >>> ukf = UKF(gyr=gyr_data, acc=acc_data, mag=mag_data)
    >>> ukf.Q.shape
    (1000, 4)

    This will perform all steps above and store the estimated orientations, as
    quaternions, in the attribute ``Q``.

    The most common sampling frequency is 100 Hz, which is used in the filter.
    If the sampling frequency is different in the given sensor data, it can be
    changed too.

    >>> ukf = UKF(gyr=gyr_data, acc=acc_data, frequency=200.0)  # Sampling frequency is 200 Hz

    The initial quaternion is estimated with the first observations of the
    tri-axial accelerometers and, if available, magnetometers, but it can also
    be given directly in the parameter ``q0``.

    >>> ukf = UKF(gyr=gyr_data, acc=acc_data, mag=mag_data, q0=[0.7071, 0.0, -0.7071, 0.0])

    Parameters
    ----------
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity in rad/s
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in mT. If not given,
        the filter will estimate orientation using only the accelerometer data.
    frequency : float, default: 100.0
        Sampling frequency in Herz.
    q0 : numpy.ndarray, default: None
        Initial orientation, as a versor (normalized quaternion).
    Dt : float, default: 0.01
        Sampling step in seconds. Inverse of sampling frequency. NOT required
        if ``frequency`` value is given.
    alpha : float, default: 1e-3
        Parameter controlling spread of Sigma points.
    beta : float, default: 2
        Parameter controlling distribution of Sigma points.
    kappa : float, default: 0
        Secondary scaling parameter. Usually set to 0.

    """
    def __init__(self,
            gyr: np.ndarray = None,
            acc: np.ndarray = None,
            mag: np.ndarray = None,
            frequency: float = 100.0,
            alpha: float = 1e-3,
            beta: float = 2,
            kappa: float = 0,
            **kwargs):
        self.gyr: np.ndarray = gyr
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.frequency: float = frequency
        self.Dt: float = kwargs.get('Dt', (1.0/self.frequency) if self.frequency else 0.01)
        self.q0: np.ndarray = kwargs.get('q0')
        # UKF parameters
        self.state_dimension: int = 4      # n : State dimension (Quaternion items)
        self.sigma_point_count: int = 2 * self.state_dimension + 1   # 2*n+1 sigma points
        self.alpha: float = alpha          # Spread parameter
        self.beta: float = beta            # Distribution parameter
        self.kappa: float = kappa          # Secondary scaling parameter
        # Lambda parameter: λ = α²(n + κ) - n
        self.lambda_param: float = self.alpha**2 * (self.state_dimension + self.kappa) - self.state_dimension
        # Weights for sigma points
        self.weight_mean, self.weight_covariance = self.set_weights()
        # Process and measurement noise covariances
        self.P: np.ndarray = kwargs.get('P', np.eye(self.state_dimension) * 0.01)    # Initial state covariance
        self.Q_t : np.ndarray = kwargs.get('process_noise_covariance', np.eye(4) * 0.0001)
        self.R : np.ndarray = kwargs.get('measurement_noise_covariance', np.eye(3) * 0.01)

        # Reference gravitational acceleration
        self.a_ref: np.ndarray = kwargs.get('a_ref', np.array([0., 0., 1.]))
        # Reference magnetic field vector
        wmm = WMM(latitude=MUNICH_LATITUDE, longitude=MUNICH_LONGITUDE, height=MUNICH_HEIGHT)
        self.m_ref = np.array([wmm.X, wmm.Y, wmm.Z])
        self.m_ref /= np.linalg.norm(self.m_ref)  # Normalize magnetic reference
        # Sensor data is given. Compute all
        if self.gyr is not None and self.acc is not None:
            self.Q: np.ndarray = self._compute_all()

    def _compute_all(self) -> np.ndarray:
        # Assert input types and values
        _assert_numerical_iterable(self.gyr, 'Angular velocity vector')
        _assert_numerical_iterable(self.acc, 'Gravitational acceleration vector')
        if self.mag is not None:
            _assert_numerical_iterable(self.mag, 'Magnetic field vector')
        self.gyr = np.array(self.gyr)
        self.acc = np.array(self.acc)
        if self.mag is not None:
            self.mag = np.array(self.mag)
        if self.acc.shape != self.gyr.shape:
            raise ValueError("acc and gyr are not the same size")
        if self.mag is not None and self.acc.shape != self.mag.shape:
            raise ValueError("acc and mag are not the same size")
        num_samples = len(self.acc)

        # Normalize sensor data
        self.acc = self.acc / np.linalg.norm(self.acc, axis=1, keepdims=True)
        if self.mag is not None:
            self.mag = self.mag / np.linalg.norm(self.mag, axis=1, keepdims=True)

        # Loop over all data
        Q = np.zeros((num_samples, 4))
        if self.mag is not None:
            # Estimation with MARG
            Q[0] = ecompass(self.acc[0], self.mag[0], frame='NED', representation='quaternion')
            for t in range(1, num_samples):
                Q[t] = self.update(q=Q[t-1], gyr=self.gyr[t], acc=self.acc[t], mag=self.mag[t])
        else:
            # Estimation with IMU
            Q[0] = acc2q(self.acc[0]) if self.q0 is None else self.q0
            Q[0] /= np.linalg.norm(Q[0])
            for t in range(1, num_samples):
                Q[t] = self.update(q=Q[t-1], gyr=self.gyr[t], acc=self.acc[t])
        return Q

    def set_weights(self) -> tuple:
        """
        Set weights for mean and covariance computation

        The weights :math:`\\mathbf{w}=\\begin{bmatrix}\\mathbf{w}^{(m)}&
        \\mathbf{w}^{(c)}\\end{bmatrix}` are computed as:

        .. math::

            \\begin{array}{rcl}
            w_0^{(m)} &=& \\frac{\\lambda}{n + \\lambda} \\\\
            w_0^{(c)} &=& \\frac{\\lambda}{n + \\lambda} + (1 - \\alpha^2 + \\beta) \\\\
            w_i^{(m)} = w_i^{(c)} &=& \\frac{1}{2(n + \\lambda)} \\quad \\text{for} \\quad i=1,2,\\ldots,2n
            \\end{array}

        The weights :math:`\\mathbf{w}^{(m)}` are used to compute the mean, and
        the weights :math:`\\mathbf{w}^{(c)}` are used to compute the
        covariance.

        The constant :math:`\\beta` is used to incorporate prior knowledge
        about the distribution of the random variable, and is usually set to
        :math:`2`.

        The scaling parameter :math:`\\alpha` determines the spread of the
        sigma points around the mean. A small value of :math:`\\alpha` results
        in sigma points that are close to the mean, while a larger value
        results in sigma points that are more spread out. The value of
        :math:`\\alpha` is usually set to a small positive number, typically in
        the range of :math:`10^{-3}` to :math:`10^{-1}`.

        Returns
        -------
        tuple
            Weights for mean and covariance.

            - ``weight_mean``: Weights for mean.
            - ``weight_covariance``: Weights for covariance.

        """
        # Weights for sigma points
        weight_mean = np.zeros(self.sigma_point_count)
        weight_covariance = np.zeros(self.sigma_point_count)
        # Set weights
        weight_mean[0] = self.lambda_param / (self.state_dimension + self.lambda_param)
        weight_covariance[0] = weight_mean[0] + (1 - self.alpha**2 + self.beta)
        weight_covariance[1:] = weight_mean[1:] = 1.0 / (2 * (self.state_dimension + self.lambda_param))
        return weight_mean, weight_covariance

    def compute_sigma_points(self, state: np.ndarray, state_covariance: np.ndarray) -> np.ndarray:
        """
        Sigma Points computation.

        Given a state :math:`\\mathbf{x}` and its covariance
        :math:`\\mathbf{P_{xx}}`, compute the sigma points
        :math:`\\mathcal{X}` as:

        .. math::

            \\begin{array}{rcl}
            \\mathcal{X}_0 &=& \\mathbf{x}_{t-1} \\\\
            \\mathcal{X}_i &=& \\mathbf{x}_{t-1} + \\mathbf{L}_i \\\\
            \\mathcal{X}_{i+n} &=& \\mathbf{x}_{t-1} - \\mathbf{L}_i
            \\end{array}

        where the **Square Root** covariance matrix :math:`\\mathbf{L}` is
        obtained using the `Cholesky decomposition
        <https://en.wikipedia.org/wiki/Cholesky_decomposition>`_:

        .. math::

            \\mathbf{L} = \\mathrm{chol}\\Big(\\sqrt{(n + \\lambda)\\mathbf{P_{xx}}}\\Big)

        Returns
        -------
        sigma_points : numpy.ndarray
            Sigma points array of shape `(2n+1, n)`, where `n` is the state
            dimension.
        """
        try:
            sqrt_covariance = np.linalg.cholesky((self.state_dimension + self.lambda_param) * abs(state_covariance))
        except np.linalg.LinAlgError:
            # Add small regularization if Cholesky decomposition fails
            regularized_covariance = state_covariance + np.eye(self.state_dimension) * 1e-8
            sqrt_covariance = np.linalg.cholesky((self.state_dimension + self.lambda_param) * regularized_covariance)
        sigma_points = np.zeros((self.state_dimension, 2 * self.state_dimension + 1))
        sigma_points[:, 0] = state
        for i in range(self.state_dimension):
            sigma_points[:, i + 1] = state + sqrt_covariance[:, i]
            sigma_points[:, self.state_dimension + i + 1] = state - sqrt_covariance[:, i]
        return sigma_points

    def Omega(self, x: np.ndarray) -> np.ndarray:
        """
        Omega operator.

        Given a vector :math:`\\mathbf{x}\\in\\mathbb{R}^3`, return a
        :math:`4\\times 4` matrix of the form:

        .. math::
            \\boldsymbol\\Omega(\\mathbf{x}) =
            \\begin{bmatrix}
                0 & -\\mathbf{x}^T \\\\
                \\mathbf{x} & -\\lfloor\\mathbf{x}\\rfloor_\\times
            \\end{bmatrix} =
            \\begin{bmatrix}
                0   & -x_1 & -x_2 & -x_3 \\\\
                x_1 &    0 &  x_3 & -x_2 \\\\
                x_2 & -x_3 &    0 &  x_1 \\\\
                x_3 &  x_2 & -x_1 & 0
            \\end{bmatrix}

        Parameters
        ----------
        x : numpy.ndarray
            Three-dimensional vector.

        Returns
        -------
        Omega : numpy.ndarray
            Omega matrix.
        """
        return np.array([
            [0.0,  -x[0], -x[1], -x[2]],
            [x[0],   0.0,  x[2], -x[1]],
            [x[1], -x[2],   0.0,  x[0]],
            [x[2],  x[1], -x[0],   0.0]])

    def update(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, mag: np.ndarray = None, dt: float = None) -> np.ndarray:
        """
        Perform an update of the state.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori state describing orientation as quaternion.
        gyr : numpy.ndarray
            Sample of tri-axial Gyroscope in rad/s.
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2.
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        updated_quaternion : numpy.ndarray
            Estimated state describing orientation as quaternion.

        """
        _assert_numerical_iterable(q, 'Quaternion')
        _assert_numerical_iterable(gyr, 'Tri-axial gyroscope sample')
        _assert_numerical_iterable(acc, 'Tri-axial accelerometer sample')
        if mag is not None:
            _assert_numerical_iterable(mag, 'Tri-axial magnetometer sample')
        dt = self.Dt if dt is None else dt

        ## Prediction Step
        # 1. Generate Sigma Points
        sigma_points = self.compute_sigma_points(q, self.P)

        # 2. Process model - propagate sigma points with gyro data
        rotation_operator = np.eye(4) + 0.5 * self.Omega(gyr) * dt
        sigma_points_propagated = np.zeros_like(sigma_points)
        for i in range(self.sigma_point_count):
            sigma_points_propagated[:, i] = Quaternion(rotation_operator @ sigma_points[:, i])

        # 3.1. Predicted state mean (y_bar)
        predicted_state_mean = Quaternion(np.sum(sigma_points_propagated * self.weight_mean, axis=1))

        # 3.2 Predicted States difference and Predicted state covariance
        # Quaternion variation using direct difference, although less robust for large rotations
        predicted_covariance = np.zeros((self.state_dimension, self.state_dimension))
        for i in range(self.sigma_point_count):
            diff = sigma_points_propagated[:, i] - predicted_state_mean
            predicted_covariance += self.weight_covariance[i] * np.outer(diff, diff)
        predicted_covariance += self.Q_t # Add process noise

        ## Correction
        # 4. Measurement Model: Transform Sigma Points into Measurement Space (expected sensor readings)
        measurement_dimension = 6 if mag is not None else 3
        sigma_points_measurement_space = np.zeros((measurement_dimension, self.sigma_point_count))
        for i in range(self.sigma_point_count):
            # Rotation matrix from inertial to body frame
            R_i_to_b = Quaternion(sigma_points_propagated[:, i]).to_DCM().T
            # Expected accelerometer reading in body frame
            sigma_points_measurement_space[:3, i] = R_i_to_b @ self.a_ref
            if mag is not None:
                # Expected magnetometer reading in body frame
                sigma_points_measurement_space[3:, i] = R_i_to_b @ self.m_ref

        # 5.1. Predicted measurement mean
        predicted_measurement_mean = np.sum(sigma_points_measurement_space * self.weight_mean, axis=1)

        # 5.2. Measurement Covariance
        measurement_covariance = np.zeros((measurement_dimension, measurement_dimension))
        # 6. Cross-Covariance
        cross_covariance = np.zeros((self.state_dimension, measurement_dimension))
        for i in range(self.sigma_point_count):
            diff_measurement = sigma_points_measurement_space[:, i] - predicted_measurement_mean
            diff_state = sigma_points_propagated[:, i] - predicted_state_mean
            measurement_covariance += self.weight_covariance[i] * np.outer(diff_measurement, diff_measurement)
            cross_covariance += self.weight_covariance[i] * np.outer(diff_state, diff_measurement)
        # Kronecker product is used to extend covariance if magnetometer is present.
        measurement_noise_covariance = np.kron(np.eye(2), self.R) if mag is not None else self.R
        measurement_covariance += measurement_noise_covariance # Add measurement noise

        # 7. Kalman Gain
        kalman_gain = cross_covariance @ np.linalg.inv(measurement_covariance)

        # 8. Innovation (measurement residual)
        actual_measurement = np.concatenate((acc, mag)) if mag is not None else acc
        innovation = actual_measurement - predicted_measurement_mean

        # 9.1. Update State Estimate
        correction = kalman_gain @ innovation
        updated_quaternion = Quaternion(predicted_state_mean + correction)

        # 9.2. Update Covariance Estimate
        self.P = predicted_covariance - kalman_gain @ measurement_covariance @ kalman_gain.T
        return updated_quaternion



# -*- coding: utf-8 -*-
"""
The Tri-Axial Attitude Determination (`TRIAD
<https://en.wikipedia.org/wiki/Triad_method>`_) was first described by Harold
Black in :cite:p:`black1964` to algebraically estimate an attitude represented
as a Direction Cosine Matrix from two orthogonal vector observations.

Given two non-parallel reference *unit vectors* :math:`\\mathbf{v}_1` and
:math:`\\mathbf{v}_2` and their corresponding *unit vectors* :math:`\\mathbf{w}_1`
and :math:`\\mathbf{w}_2`, it is required to find an orthogonal matrix
:math:`\\mathbf{A}` satisfying:

.. math::
    \\mathbf{Av}_1

Two vectors :math:`\\mathbf{v}_1` and :math:`\\mathbf{v}_2` define an
orthogonal coordinate system with the **normalized** basis vectors
:math:`\\mathbf{q}`, :math:`\\mathbf{r}`, and :math:`\\mathbf{s}` as the
following triad:

.. math::
    \\begin{array}{rcl}
    \\mathbf{q}_r &=& \\mathbf{v}_1 \\\\
    \\mathbf{r}_r &=& \\frac{\\mathbf{v}_1\\times\\mathbf{v}_2}{|\\mathbf{v}_1\\times\\mathbf{v}_2|} \\\\
    \\mathbf{s}_r &=& \\mathbf{q}_r\\times\\mathbf{r}_r
    \\end{array}

The TRIAD method, initially developed to estimate the attitude of spacecrafts
:cite:p:`shuster2007`, uses the position of the sun (using a `star tracker
<https://en.wikipedia.org/wiki/Star_tracker>`_) and the magnetic field of Earth
as references :cite:p:`hall2000` :cite:p:`markley2014`. These are represented
as vectors to build an appropriate *reference* frame :math:`\\mathbf{M}_r`:

.. math::
    \\mathbf{M}_r = \\begin{bmatrix} \\mathbf{q}_r & \\mathbf{r}_r & \\mathbf{s}_r \\end{bmatrix}

Similarly, at any given time, two measured vectors in the spacecraft's **body
frame** :math:`\\mathbf{w}_1` and :math:`\\mathbf{w}_2` determine the
:math:`3\\times 3` body matrix :math:`\\mathbf{M}_b`:

.. math::
    \\mathbf{M}_b = \\begin{bmatrix} \\mathbf{q}_b & \\mathbf{r}_b & \\mathbf{s}_b \\end{bmatrix}

where, like the first triad, the second triad is built as:

.. math::
    \\begin{array}{rcl}
    \\mathbf{q}_b &=& \\mathbf{w}_1 \\\\
    \\mathbf{r}_b &=& \\frac{\\mathbf{w}_1\\times\\mathbf{w}_2}{|\\mathbf{w}_1\\times\\mathbf{w}_2|} \\\\
    \\mathbf{s}_b &=& \\mathbf{q}_b\\times\\mathbf{r}_b
    \\end{array}

The attitude matrix :math:`\\mathbf{A}\\in\\mathbb{R}^{3\\times 3}` defines the
coordinate transformation,

.. math::
    \\mathbf{AM}_r = \\mathbf{M}_b

Solving for :math:`\\mathbf{A}` we obtain:

.. math::
    \\mathbf{A} = \\mathbf{M}_b\\mathbf{M}_r^{-1}

But we also know that :math:`\\mathbf{M}_r` is orthogonal. So, the solution is
simply:

.. math::
    \\mathbf{A} = \\mathbf{M}_b\\mathbf{M}_r^T

Inverse trigonometric functions are not required, a unique attitude is obtained,
and computational requirements are minimal.

It is only required that :math:`\\mathbf{M}_r` has an inverse, but that is
already ensured, since :math:`\\mathbf{q}_r`, :math:`\\mathbf{r}_r`,
and :math:`\\mathbf{s}_r` are linearly independent :cite:p:`lerner1978`.

Strapdown INS
-------------

For estimations using a Strapdown INS on Earth, we identify two main reference
vectors: gravity :math:`\\mathbf{g}=\\begin{bmatrix}g_x & g_y & g_z\\end{bmatrix}`
and magnetic field :math:`\\mathbf{h}=\\begin{bmatrix}h_x & h_y & h_z\\end{bmatrix}`.

A common convention sets the *gravity vector* equal to :math:`0` along the X-
and Y-axis, and equal to :math:`\\sim 9.81` along the Z-axis. This assumes the
direction of the gravity is parallel to the vertical axis. Because TRIAD uses
normalized vectors, the Z-axis will turn out to be equal to :math:`1`:

.. math::
    \\mathbf{g} = \\begin{bmatrix}0 \\\\ 0 \\\\ 1 \\end{bmatrix}

The *magnetic field* is defined from the geographical position of the
measurement. Using the `World Magnetic Model <https://www.ngdc.noaa.gov/geomag/WMM/>`_,
we can estimate the magnetic field elements of our location on a given date.

The class :class:`ahrs.utils.WMM` can help us to retrieve it. Let's say we want
to know the geomagnetic field elements of Munich, Germany [#]_ on the 3rd of
October, 2020.

The city's location is 48.137154° N and 11.576124° E at 519 m above sea level.
We obtain its magnetic elements as:

.. code-block:: python

    >>> import datetime
    >>> from ahrs.utils import WMM
    >>> wmm = WMM(latitude=48.137154, longitude=11.576124, height=0.519, date=datetime.date(2020, 10, 3))
    >>> wmm.magnetic_elements
    {'X': 21009.66924050522, 'Y': 1333.4601319284525, 'Z': 43731.849938722924, 'H': 21051.943319296533, 'F': 48535.13177670226, 'I': 64.2944417667441, 'D': 3.631627635223863, 'GV': 3.631627635223863}

For further explanation of class :class:`WMM`, please check its `page <../WMM.html>`_.
Of our interest are only the values of ``X``, ``Y`` and ``Z`` representing the
magnetic field intensity, in nT, along the X-, Y- and Z-axis, respectively.

.. math::
    \\mathbf{h} = \\begin{bmatrix} 21009.66924 \\\\ 1333.46013 \\\\ 43731.84994 \\end{bmatrix}

But, again, TRIAD works with normalized vectors, so the reference magnetic
vector becomes:

.. math::
    \\mathbf{h} = \\begin{bmatrix} 0.43288 \\\\ 0.02747 \\\\ 0.90103 \\end{bmatrix}

.. code-block:: python

    >>> import numpy as np
    >>> h = np.array([wmm.magnetic_elements[x] for x in list('XYZ')])
    >>> h /= np.linalg.norm(h)      # Reference geomagnetic field (h)
    >>> h
    array([0.4328755 , 0.02747412, 0.90103495])

Both normalized vectors :math:`\\mathbf{g}` and :math:`\\mathbf{h}` build the
*reference triad* :math:`\\mathbf{M}_r`

Then, we have to measure their equivalent vectors, for which we use the
accelerometer to obtain :math:`\\mathbf{a} = \\begin{bmatrix}a_x & a_y & a_z \\end{bmatrix}`,
and the magnetometer for :math:`\\mathbf{m} = \\begin{bmatrix}m_x & m_y & m_z \\end{bmatrix}`.

Both measurement vectors are also normalized, meaning :math:`\\|\\mathbf{a}\\|=\\|\\mathbf{m}\\|=1`,
so that they can build the *body's measurement triad*  :math:`\\mathbf{M}_b`.

To get the Direction Cosine Matrix we simply call the method ``estimate`` with
the normalized measurement vectors:

.. code-block:: python

    >>> triad = ahrs.filters.TRIAD()
    >>> triad.v1 = np.array([0.0, 0.0, 1.0])                    # Reference gravity vector (g)
    >>> triad.v2 = h                                            # Reference geomagnetic field (h)
    >>> a = np.array([-2.499e-04, 4.739e-02, 0.9988763])        # Measured acceleration (normalized)
    >>> a /= np.linalg.norm(a)
    >>> m = np.array([-0.36663061, 0.17598138, -0.91357132])    # Measured magnetic field (normalized)
    >>> m /= np.linalg.norm(m)
    >>> triad.estimate(w1=a, w2=m)
    array([[-8.48320410e-01, -5.29483162e-01, -2.49900033e-04],
           [ 5.28878238e-01, -8.47373587e-01,  4.73900062e-02],
           [-2.53039690e-02,  4.00697428e-02,  9.98876431e-01]])

Optionally, it can return the estimation as a quaternion representation setting
``representation`` to ``'quaternion'``.

.. code-block:: python

    >>> triad.estimate(w1=a, w2=m, representation='quaternion')
    array([ 0.27531002, -0.00664729,  0.02275078,  0.96106327])

Giving the observation vector to the constructor, the attitude estimation
happens automatically, and is stored in the attribute ``A``.

.. code-block:: python

    >>> triad = ahrs.filters.TRIAD(w1=np.array([-2.499e-04, 4.739e-02, 0.9988763]), w2=np.array([-0.36663061, 0.17598138, -0.91357132]), v2=h)
    >>> triad.A
    array([[-8.48320410e-01, -5.29483162e-01, -2.49900033e-04],
           [ 5.28878238e-01, -8.47373587e-01,  4.73900062e-02],
           [-2.53039690e-02,  4.00697428e-02,  9.98876431e-01]])
    >>> triad = ahrs.filters.TRIAD(w1=np.array([-2.499e-04, 4.739e-02, 0.9988763]), w2=np.array([-0.36663061, 0.17598138, -0.91357132]), v2=h, representation='quaternion')
    >>> triad.A
    array([ 0.27531002, -0.00664729,  0.02275078,  0.96106327])

If the input data contains many observations, all will be estimated at once.

.. code-block:: python

    >>> a = np.array([[-0.000249905733, 0.0473926177, 0.998876307],
    ... [-0.00480145530, 0.0572267567, 0.998349660],
    ... [-0.00986626329, 0.0746539896, 0.997160688]])
    >>> m = np.array([[-0.36663061, 0.17598138, -0.91357132],
    ... [-0.37726367, 0.18069746, -0.90830642],
    ... [-0.3874741, 0.18536454, -0.9030525]])
    >>> triad = ahrs.filters.TRIAD(w1=a, w2=m, v2=h)
    >>> triad.A
    array([[[-8.48317898e-01, -5.29487187e-01, -2.49905733e-04],
            [ 5.28882192e-01, -8.47370974e-01,  4.73926177e-02],
            [-2.53055467e-02,  4.00718352e-02,  9.98876307e-01]],

           [[-8.43678607e-01, -5.36827117e-01, -4.80145530e-03],
            [ 5.35721702e-01, -8.42453178e-01,  5.72267567e-02],
            [-3.47658761e-02,  4.57087466e-02,  9.98349660e-01]],

           [[-8.32771974e-01, -5.53528225e-01, -9.86626329e-03],
            [ 5.51396878e-01, -8.30896061e-01,  7.46539896e-02],
            [-4.95209297e-02,  5.67295235e-02,  9.97160688e-01]]])
    >>> triad = ahrs.filters.TRIAD(w1=a, w2=m, representation='quaternion')
    >>> triad.A
    array([[ 0.27531229, -0.00664771,  0.02275202,  0.96106259],
           [ 0.2793823 , -0.01030667,  0.0268131 ,  0.95975016],
           [ 0.28874411, -0.01551933,  0.03433374,  0.95666461]])

The first disadvantage is that TRIAD can only use two observations per
estimation. If there are more observations, we must discard part of them
(losing accuracy), or mix them in such a way that we obtain only two
representative observations.

The second disadvantage is its loss of accuracy in a heavily dynamic state of
the measuring device. TRIAD assumes a quasi-static state of the body frame and,
therefore, its use is limited to motionless objects, preferably.

Footnotes
---------
.. [#] This package's author resides in Munich, and examples of geographical
    locations will take it as a reference.

"""

import numpy as np
from ..common.orientation import chiaverini
from ..common.constants import MUNICH_LATITUDE
from ..common.constants import MUNICH_LONGITUDE
from ..common.constants import MUNICH_HEIGHT
from ..common.mathfuncs import cosd
from ..common.mathfuncs import sind

class TRIAD:
    """
    Tri-Axial Attitude Determination

    TRIAD estimates the attitude as a Direction Cosine Matrix. To return it as
    a quaternion, set the parameter ``representation`` to ``'quaternion'``.

    Parameters
    ----------
    w1 : numpy.ndarray
        First tri-axial observation vector in body frame. Usually a normalized
        acceleration vector :math:`\\mathbf{a} = \\begin{bmatrix}a_x & a_y & a_z \\end{bmatrix}`
    w2 : numpy.ndarray
        Second tri-axial observation vector in body frame. Usually a normalized
        magnetic field vector :math:`\\mathbf{m} = \\begin{bmatrix}m_x & m_y & m_z \\end{bmatrix}`
    v1 : numpy.ndarray, optional.
        First tri-axial reference vector. Defaults to normalized gravity vector
        :math:`\\mathbf{g} = \\begin{bmatrix}0 & 0 & 1 \\end{bmatrix}`
    v2 : numpy.ndarray, optional.
        Second tri-axial reference vector. Defaults to normalized geomagnetic
        field :math:`\\mathbf{h} = \\begin{bmatrix}h_x & h_y & h_z \\end{bmatrix}`
        in Munich, Germany.
    representation : str, default: ``'rotmat'``
        Attitude representation. Options are ``'rotmat'`` or ``'quaternion'``.
    frame : str, default: ``'NED'``
        Local tangent plane coordinate frame. Valid options are right-handed
        ``'NED'`` for North-East-Down and ``'ENU'`` for East-North-Up.

    Attributes
    ----------
    w1 : numpy.ndarray
        First tri-axial observation vector in body frame.
    w2 : numpy.ndarray
        Second tri-axial observation vector in body frame.
    v1 : numpy.ndarray, optional.
        First tri-axial reference vector.
    v2 : numpy.ndarray, optional.
        Second tri-axial reference vector.
    A : numpy.ndarray
        Estimated attitude.

    Examples
    --------
    >>> from ahrs.filters import TRIAD
    >>> triad = TRIAD()
    >>> triad.v1 = np.array([0.0, 0.0, 1.0])                    # Reference gravity vector (g)
    >>> triad.v2 = np.array([21.0097, 1.3335, 43.732])          # Reference geomagnetic field (h)
    >>> a = np.array([-2.499e-04, 4.739e-02, 0.9988763])        # Measured acceleration (normalized)
    >>> a /= np.linalg.norm(a)
    >>> m = np.array([-0.36663061, 0.17598138, -0.91357132])    # Measured magnetic field (normalized)
    >>> m /= np.linalg.norm(m)
    >>> triad.estimate(w1=a, w2=m)
    array([[-8.48320410e-01, -5.29483162e-01, -2.49900033e-04],
           [ 5.28878238e-01, -8.47373587e-01,  4.73900062e-02],
           [-2.53039690e-02,  4.00697428e-02,  9.98876431e-01]])

    It also works by passing each array to its corresponding parameter. They
    will be normalized too.

    >>> triad = TRIAD(w1=a, w2=m, v1=[0.0, 0.0, 1.0], v2=[-0.36663061, 0.17598138, -0.91357132])

    """
    def __init__(self,
        w1: np.ndarray = None,
        w2: np.ndarray = None,
        v1: np.ndarray = None,
        v2: np.ndarray = None,
        representation: str = 'rotmat',
        frame: str = 'NED'):
        self.representation: str = representation
        self.frame: str = frame
        self._guard_clauses_parameters()
        self._guard_clauses_vectors(w1, w2, v1, v2)
        # Input values
        self.w1: np.ndarray = np.copy(w1) if isinstance(w1, (list, np.ndarray)) else None
        self.w2: np.ndarray = np.copy(w2) if isinstance(w2, (list, np.ndarray)) else None
        # Reference frames
        self.v1: np.ndarray = self._set_first_triad_reference(v1)
        self.v2: np.ndarray = self._set_second_triad_reference(v2)
        # Compute values if samples given
        if self.w1 is not None and self.w2 is not None:
            self.A: np.ndarray = self._compute_all(self.representation)

    def _guard_clauses_parameters(self) -> None:
        for item in [self.representation, self.frame]:
            if not isinstance(item, str):
                raise TypeError(f"{item} must be a string. Got {type(item)}")
        if self.representation.lower() not in ['rotmat', 'quaternion']:
            raise ValueError(f"Given representation '{self.representation}' is NOT valid. Try 'rotmat', or 'quaternion'")
        if self.frame.upper() not in ['NED', 'ENU']:
            raise ValueError(f"Given frame '{self.frame}' is NOT valid. Try 'NED' or 'ENU'")

    def _guard_clauses_vectors(self, *vectors) -> None:
        for item in vectors:
            if not isinstance(item, (list, np.ndarray, type(None))):
                raise TypeError(f"Vectors must be given in a list or numpy.ndarray. It is {type(item)}")
            if isinstance(item, type(None)):
                continue
            if np.linalg.norm(item) == 0:
                raise ValueError("Vector must be non-zero.")
            item_copy = np.copy(item)
            if item_copy.ndim < 1 or item_copy.ndim > 2:
                raise ValueError(f"Vector must be 1- or 2-dimensional. It is {item_copy.ndim}-dimensional.")
            if item_copy.shape[-1] != 3:
                raise ValueError(f"Vector must have 3 items. It has {item_copy.shape[-1]}.")
        w1, w2, *_ = vectors
        if w1 is None or w2 is None:
            return None
        w1, w2 = np.copy(w1), np.copy(w2)
        if w1.shape != w2.shape:
            raise ValueError(f"Vectors must have the same shape. w1: {w1.shape}, w2: {w2.shape}")

    def _set_first_triad_reference(self, vector: np.ndarray) -> np.ndarray:
        if vector is None:
            ref = np.array([0.0, 0.0, 1.0]) if self.frame.upper() == 'NED' else np.array([0.0, 0.0, -1.0])
            return ref
        return np.copy(vector)/np.linalg.norm(vector)

    def _set_second_triad_reference(self, value: np.ndarray) -> np.ndarray:
        if isinstance(value, float):
            if abs(value) > 90:
                raise ValueError(f"Dip Angle must be within range [-90, 90]. Got {value}")
            ref = np.array([cosd(value), 0.0, sind(value)]) if self.frame.upper() == 'NED' else np.array([0.0, cosd(value), -sind(value)])
        elif isinstance(value, (np.ndarray, list)):
            ref = np.copy(value)
        else:
            # Reference Observations in Munich, Germany
            from ..utils.wmm import WMM
            mag = WMM(latitude=MUNICH_LATITUDE, longitude=MUNICH_LONGITUDE, height=MUNICH_HEIGHT).magnetic_elements
            ref = np.array([mag['X'], mag['Y'], mag['Z']])
        return ref/np.linalg.norm(ref)

    def _compute_all(self, representation: str) -> np.ndarray:
        """
        Estimate the attitude given all data.

        Attributes ``w1`` and ``w2`` must contain data.

        Parameters
        ----------
        representation : str
            Attitude representation. Options are ``'rotmat'`` or ``'quaternion'``.

        Returns
        -------
        A : numpy.ndarray
            M-by-3-by-3 with all estimated attitudes as direction cosine
            matrices, where M is the number of samples. It is an N-by-4 array
            if ``representation`` is set to ``'quaternion'``.

        """
        if self.w1.shape != self.w2.shape:
            raise ValueError("w1 and w2 are not the same size")
        if self.w1.ndim == 1:
            return self.estimate(self.w1, self.w2, representation)
        num_samples = len(self.w1)
        A = np.zeros((num_samples, 4)) if representation.lower() == 'quaternion' else np.zeros((num_samples, 3, 3))
        for t in range(num_samples):
            A[t] = self.estimate(self.w1[t], self.w2[t], representation)
        return A

    def estimate(self, w1: np.ndarray, w2: np.ndarray, representation: str = 'rotmat') -> np.ndarray:
        """
        Attitude Estimation.

        The equation numbers in the code refer to :cite:p:`lerner1978`.

        Parameters
        ----------
        w1 : numpy.ndarray
            Sample of first tri-axial sensor.
        w2 : numpy.ndarray
            Sample of second tri-axial sensor.
        representation : str, default: ``'rotmat'``
            Attitude representation. Options are ``rotmat'`` or ``'quaternion'``.

        Returns
        -------
        A : numpy.ndarray
            Estimated attitude as 3-by-3 Direction Cosine Matrix. If
            ``representation`` is set to ``'quaternion'``, it is returned as a
            quaternion.

        Examples
        --------
        >>> triad = ahrs.filters.TRIAD()
        >>> triad.v1 = [0.0, 0.0, 1.0]                              # Normalized reference gravity vector (g)
        >>> triad.v2 = [0.4328755, 0.02747412, 0.90103495]          # Normalized reference geomagnetic field (h)
        >>> a = [4.098297, 8.663757, 2.1355896]                     # Measured acceleration
        >>> m = [-28715.50512, -25927.43566, 4756.83931]            # Measured magnetic field
        >>> triad.estimate(w1=a, w2=m)                              # Estimated attitude as DCM
        array([[-7.84261e-01  ,  4.5905718e-01,  4.1737417e-01],
               [ 2.2883429e-01, -4.1126404e-01,  8.8232463e-01],
               [ 5.7668844e-01,  7.8748232e-01,  2.1749032e-01]])
        >>> triad.estimate(w1=a, w2=m, representation='quaternion')          # Estimated attitude as quaternion
        array([ 0.07410345, -0.3199659, -0.53747247, -0.77669417])

        """
        if representation.lower() not in ['rotmat', 'quaternion']:
            raise ValueError(f"Given representation '{representation}' is NOT valid. Try 'rotmat', or 'quaternion'")
        w1, w2 = np.copy(w1), np.copy(w2)
        # Normalized Vectors
        w1 /= np.linalg.norm(w1)                            # (eq. 12-39a)
        w2 /= np.linalg.norm(w2)
        # First Triad
        w1xw2 = np.cross(w1, w2)
        s2 = w1xw2 / np.linalg.norm(w1xw2)                  # (eq. 12-39b)
        s3 = np.cross(w1, w1xw2) / np.linalg.norm(w1xw2)    # (eq. 12-39c)
        # Second Triad
        v1xv2 = np.cross(self.v1, self.v2)
        r2 = v1xv2 / np.linalg.norm(v1xv2)
        r3 = np.cross(self.v1, v1xv2) / np.linalg.norm(v1xv2)
        # Solve TRIAD
        Mb = np.c_[w1, s2, s3]                              # (eq. 12-41)
        Mr = np.c_[self.v1, r2, r3]                         # (eq. 12-42)
        A = Mb@Mr.T                                         # (eq. 12-45)
        # Return according to desired representation
        if representation.lower() == 'quaternion':
            return chiaverini(A)
        return A



# -*- coding: utf-8 -*-
"""
This is a modified `OLEQ <./oleq.html>`_, where a recursive estimation of the
attitude is made with the measured angular velocity :cite:p:`zhou2018`. This
estimation is set as the initial value for the OLEQ estimation, simplyfing the
rotational operations.

First, the quaternion :math:`\\mathbf{q}_\\omega` is estimated from the angular
velocity, :math:`\\boldsymbol\\omega=\\begin{bmatrix}\\omega_x & \\omega_y &
\\omega_z \\end{bmatrix}^T`, measured by the gyroscopes, in rad/s, at a time
:math:`t` as:

.. math::
    \\mathbf{q}_\\omega = \\Big(\\mathbf{I}_4 + \\frac{\\Delta t}{2}\\boldsymbol\\Omega_t\\Big)\\mathbf{q}_{t-1} =
    \\begin{bmatrix}
    q_w - \\frac{\\Delta t}{2} \\omega_x q_x - \\frac{\\Delta t}{2} \\omega_y q_y - \\frac{\\Delta t}{2} \\omega_z q_z\\\\
    q_x + \\frac{\\Delta t}{2} \\omega_x q_w - \\frac{\\Delta t}{2} \\omega_y q_z + \\frac{\\Delta t}{2} \\omega_z q_y\\\\
    q_y + \\frac{\\Delta t}{2} \\omega_x q_z + \\frac{\\Delta t}{2} \\omega_y q_w - \\frac{\\Delta t}{2} \\omega_z q_x\\\\
    q_z - \\frac{\\Delta t}{2} \\omega_x q_y + \\frac{\\Delta t}{2} \\omega_y q_x + \\frac{\\Delta t}{2} \\omega_z q_w
    \\end{bmatrix}

Then, the attitude is "corrected" through OLEQ using a single multiplication of
its rotation operator:

.. math::
    \\mathbf{q}_\\mathbf{ROLEQ} = \\frac{1}{2}\\Big(\\mathbf{I}_4 + \\sum_{i=1}^na_i\\mathbf{W}_i\\Big)\\mathbf{q}_\\omega

where each :math:`\\mathbf{W}` (one for accelerations and one for magnetic
field) is built from their reference vectors, :math:`D^r`, and measurements,
:math:`D^b`, exactly as in OLEQ:

.. math::
    \\begin{array}{rcl}
    \\mathbf{W} &=& D_x^r\\mathbf{M}_1 + D_y^r\\mathbf{M}_2 + D_z^r\\mathbf{M}_3 \\\\ && \\\\
    \\mathbf{M}_1 &=&
    \\begin{bmatrix}
    D_x^b & 0 & D_z^b & -D_y^b \\\\
    0 & D_x^b & D_y^b & D_z^b \\\\
    D_z^b & D_y^b & -D_x^b & 0 \\\\
    -D_y^b & D_z^b & 0 & -D_x^b
    \\end{bmatrix} \\\\
    \\mathbf{M}_2 &=&
    \\begin{bmatrix}
    D_y^b & -D_z^b & 0 & D_x^b \\\\
    -D_z^b & -D_y^b & D_x^b & 0 \\\\
    0 & D_x^b & D_y^b & D_z^b \\\\
    D_x^b & 0 & D_z^b & -D_y^b
    \\end{bmatrix} \\\\
    \\mathbf{M}_3 &=&
    \\begin{bmatrix}
    D_z^b & D_y^b & -D_x^b & 0 \\\\
    D_y^b & -D_z^b & 0 & D_x^b \\\\
    -D_x^b & 0 & -D_z^b & D_y^b \\\\
    0 & D_x^b & D_y^b & D_z^b
    \\end{bmatrix}
    \\end{array}

It is noticeable that, for OLEQ, a random quaternion was used as a starting
value for an iterative procedure to find the optimal quaternion. Here, that
initial value is now :math:`\\mathbf{q}_\\omega` and a simple product (instead
of a large iterative product) is required.

In this way, the quaternions are recursively computed with much fewer
computations, and the accuracy is maintained.

For this case, however the three sensor data (gyroscopes, accelerometers and
magnetometers) have to be provided, along with the an initial quaternion,
:math:`\\mathbf{q}_0` from which the attitude will be built upon.

"""

import numpy as np
from . import OLEQ
from ..common.mathfuncs import cosd
from ..common.mathfuncs import sind
from ..utils.core import _assert_numerical_iterable
from ..utils.core import _assert_numerical_positive_variable

class ROLEQ:
    """
    Recursive Optimal Linear Estimator of Quaternion

    Uses OLEQ to estimate the initial attitude.

    Parameters
    ----------
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity in rad/s.
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2.
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in mT.

    Attributes
    ----------
    gyr : numpy.ndarray
        N-by-3 array with N gyroscope samples.
    acc : numpy.ndarray
        N-by-3 array with N accelerometer samples.
    mag : numpy.ndarray
        N-by-3 array with N magnetometer samples.
    frequency : float
        Sampling frequency in Herz
    Dt : float
        Sampling step in seconds. Inverse of sampling frequency.
    Q : numpy.array, default: None
        M-by-4 Array with all estimated quaternions, where M is the number of
        samples. Equal to None when no estimation is performed.

    Raises
    ------
    ValueError
        When dimension of input arrays ``gyr``, ``acc`` or ``mag`` are not
        equal.

    Examples
    --------
    >>> gyr_data.shape, acc_data.shape, mag_data.shape      # NumPy arrays with sensor data
    ((1000, 3), (1000, 3), (1000, 3))
    >>> from ahrs.filters import ROLEQ
    >>> orientation = ROLEQ(gyr=gyr_data, acc=acc_data, mag=mag_data)
    >>> orientation.Q.shape                 # Estimated attitude
    (1000, 4)

    """
    def __init__(self,
        gyr: np.ndarray = None,
        acc: np.ndarray = None,
        mag: np.ndarray = None,
        weights: np.ndarray = None,
        magnetic_ref: np.ndarray = None,
        frame: str = 'NED',
        **kwargs
        ):
        self.gyr: np.ndarray = gyr
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.a: np.ndarray = weights if weights is not None else np.ones(2)
        self.frequency: float = kwargs.get('frequency', 100.0)
        self.Dt: float = kwargs.get('Dt', (1.0/self.frequency) if self.frequency else 0.01)
        self.q0: np.ndarray = kwargs.get('q0')
        self.frame: str = frame
        # Reference measurements
        self._set_reference_frames(magnetic_ref, self.frame)
        self._assert_validity_of_inputs()
        # Estimate all quaternions if data is given
        if self.acc is not None and self.gyr is not None and self.mag is not None:
            self.Q = self._compute_all()

    def _set_reference_frames(self, mref: float, frame: str = 'NED'):
        if not isinstance(frame, str):
            raise TypeError(f"'frame' must be a string. Got {type(frame)}.")
        if frame.upper() not in ['NED', 'ENU']:
            raise ValueError(f"Invalid frame '{frame}'. Try 'NED' or 'ENU'")
        #### Magnetic Reference Vector ####
        if mref is None:
            # Local magnetic reference of Munich, Germany
            from ..common.constants import MUNICH_LATITUDE, MUNICH_LONGITUDE, MUNICH_HEIGHT
            from ..utils.wmm import WMM
            wmm = WMM(latitude=MUNICH_LATITUDE, longitude=MUNICH_LONGITUDE, height=MUNICH_HEIGHT)
            cd, sd = cosd(wmm.I), sind(wmm.I)
            self.m_ref = np.array([sd, 0.0, cd]) if frame.upper() == 'NED' else np.array([0.0, cd, -sd])
        elif isinstance(mref, bool):
            raise TypeError(f"'mref' must be a float or numpy.ndarray. Got {type(mref)}.")
        elif isinstance(mref, (int, float)):
            # Use given magnetic dip angle (in degrees)
            cd, sd = cosd(mref), sind(mref)
            self.m_ref = np.array([sd, 0.0, cd]) if frame.upper() == 'NED' else np.array([0.0, cd, -sd])
        elif isinstance(mref, (list, tuple, np.ndarray)):
            # Magnetic reference is given as a vector
            self.m_ref = np.copy(mref)
        else:
            raise TypeError(f"Magnetic reference must be float, int, list, tuple or numpy.ndarray. Got {type(mref)}.")
        if self.m_ref.shape != (3,):
            raise ValueError(f"Magnetic reference vector must be of shape (3,). Got {self.m_ref.shape}.")
        if not any(self.m_ref):
            raise ValueError("Magnetic reference vector must contain non-zero values.")
        self.m_ref /= np.linalg.norm(self.m_ref)
        #### Gravitational Reference Vector ####
        self.a_ref = np.array([0.0, 0.0, -1.0]) if frame.upper() == 'NED' else np.array([0.0, 0.0, 1.0])

    def _assert_validity_of_inputs(self):
        """Asserts the validity of the inputs."""
        # Assert floats
        for item in ["frequency", "Dt"]:
            _assert_numerical_positive_variable(getattr(self, item), item)
        # Assert arrays
        for item in ['gyr', 'acc', 'mag', 'a', 'm_ref', 'a_ref', 'q0']:
            if self.__getattribute__(item) is not None:
                if isinstance(self.__getattribute__(item), bool):
                    raise TypeError(f"Parameter '{item}' must be an array of numeric values.")
                _assert_numerical_iterable(self.__getattribute__(item), item)
                self.__setattr__(item, np.copy(self.__getattribute__(item)))
        if self.acc is not None and self.mag is None:
            raise ValueError("If 'acc' is given, 'mag' must also be given.")
        if self.mag is not None and self.acc is None:
            raise ValueError("If 'mag' is given, 'acc' must also be given.")
        if self.q0 is not None:
            if self.q0.shape != (4,):
                raise ValueError(f"Parameter 'q0' must be an array of shape (4,). It is {self.q0.shape}.")
            if not np.allclose(np.linalg.norm(self.q0), 1.0):
                raise ValueError(f"Parameter 'q0' must be a versor (norm equal to 1.0). Its norm is equal to {np.linalg.norm(self.q0)}.")
        # Assert weights
        if self.a.shape != (2,):
            raise ValueError(f"Dimension of 'weights' must be (2,). Got {self.a.shape}.")
        for item in self.a:
            if not isinstance(item, (int, float)):
                raise TypeError(f"'weights' must be an array of numeric values. Got {type(item)}.")
            if item < 0.0:
                raise ValueError(f"'weights' must be non-negative. Got {item}.")
        if not any(self.a > 0):
            raise ValueError("'weights' must contain positive values.")

    def _compute_all(self) -> np.ndarray:
        """
        Estimate the quaternions given all data.

        Attributes ``gyr``, ``acc`` and ``mag`` must contain data.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        if self.acc.shape != self.gyr.shape:
            raise ValueError("acc and gyr are not the same size")
        if self.acc.shape != self.mag.shape:
            raise ValueError("acc and mag are not the same size")
        num_samples = np.atleast_2d(self.acc).shape[0]
        if num_samples < 2:
            raise ValueError("ROLEQ needs at least 2 samples of each sensor")
        Q = np.zeros((num_samples, 4))
        oleq = OLEQ(magnetic_ref=self.m_ref, frame=self.frame)
        Q[0] = oleq.estimate(self.acc[0], self.mag[0]) if self.q0 is None else self.q0
        for t in range(1, num_samples):
            Q[t] = self.update(Q[t-1], self.gyr[t], self.acc[t], self.mag[t])
        return Q

    def attitude_propagation(self, q: np.ndarray, omega: np.ndarray, dt: float) -> np.ndarray:
        """
        Attitude estimation from previous quaternion and current angular velocity.

        .. math::
            \\mathbf{q}_\\omega = \\Big(\\mathbf{I}_4 + \\frac{\\Delta t}{2}\\boldsymbol\\Omega_t\\Big)\\mathbf{q}_{t-1} =
            \\begin{bmatrix}
            q_w - \\frac{\\Delta t}{2} \\omega_x q_x - \\frac{\\Delta t}{2} \\omega_y q_y - \\frac{\\Delta t}{2} \\omega_z q_z\\\\
            q_x + \\frac{\\Delta t}{2} \\omega_x q_w - \\frac{\\Delta t}{2} \\omega_y q_z + \\frac{\\Delta t}{2} \\omega_z q_y\\\\
            q_y + \\frac{\\Delta t}{2} \\omega_x q_z + \\frac{\\Delta t}{2} \\omega_y q_w - \\frac{\\Delta t}{2} \\omega_z q_x\\\\
            q_z - \\frac{\\Delta t}{2} \\omega_x q_y + \\frac{\\Delta t}{2} \\omega_y q_x + \\frac{\\Delta t}{2} \\omega_z q_w
            \\end{bmatrix}

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        omega : numpy.ndarray
            Angular velocity, in rad/s.
        dt : float
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Attitude as a quaternion.
        """
        Omega_t = np.array([
            [0.0,  -omega[0], -omega[1], -omega[2]],
            [omega[0],   0.0,  omega[2], -omega[1]],
            [omega[1], -omega[2],   0.0,  omega[0]],
            [omega[2],  omega[1], -omega[0],   0.0]])
        q_omega = (np.identity(4) + 0.5*dt*Omega_t) @ q    # (eq. 37)
        return q_omega/np.linalg.norm(q_omega)

    def WW(self, Db, Dr):
        """
        W Matrix

        .. math::
            \\mathbf{W} = D_x^r\\mathbf{M}_1 + D_y^r\\mathbf{M}_2 + D_z^r\\mathbf{M}_3

        Parameters
        ----------
        Db : numpy.ndarray
            Normalized tri-axial observations vector.
        Dr : numpy.ndarray
            Normalized tri-axial reference vector.

        Returns
        -------
        W_matrix : numpy.ndarray
            W Matrix.
        """
        bx, by, bz = Db
        rx, ry, rz = Dr
        M1 = np.array([
            [bx, 0.0, bz, -by],
            [0.0, bx, by, bz],
            [bz, by, -bx, 0.0],
            [-by, bz, 0.0, -bx]])       # (eq. 18a)
        M2 = np.array([
            [by, -bz, 0.0, bx],
            [-bz, -by, bx, 0.0],
            [0.0, bx, by, bz],
            [bx, 0.0, bz, -by]])        # (eq. 18b)
        M3 = np.array([
            [bz, by, -bx, 0.0],
            [by, -bz, 0.0, bx],
            [-bx, 0.0, -bz, by],
            [0.0, bx, by, bz]])         # (eq. 18c)
        return rx*M1 + ry*M2 + rz*M3    # (eq. 20)

    def oleq(self, acc: np.ndarray, mag: np.ndarray, q_omega: np.ndarray) -> np.ndarray:
        """
        OLEQ with a single rotation by R.

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer.
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer.
        q_omega : numpy.ndarray
            Preceding quaternion estimated with angular velocity.

        Returns
        -------
        q : numpy.ndarray
            Final quaternion.

        """
        _assert_numerical_iterable(acc, 'Gravitational acceleration vector')
        _assert_numerical_iterable(mag, 'Geomagnetic field vector')
        a_norm = np.linalg.norm(acc)
        m_norm = np.linalg.norm(mag)
        if not a_norm > 0 or not m_norm > 0:    # handle NaN
            return q_omega
        acc = np.copy(acc) / np.linalg.norm(acc)
        mag = np.copy(mag) / np.linalg.norm(mag)
        sum_aW = self.a[0]*self.WW(acc, self.a_ref) + self.a[1]*self.WW(mag, self.m_ref)    # (eq. 31)
        R = 0.5*(np.identity(4) + sum_aW)       # (eq. 33)
        q = R @ q_omega                         # (eq. 25)
        return q / np.linalg.norm(q)

    def update(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, mag: np.ndarray, dt: float = None) -> np.ndarray:
        """
        Update Attitude with a Recursive OLEQ

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Sample of angular velocity in rad/s
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer in mT
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        """
        dt = self.Dt if dt is None else dt
        q_g = self.attitude_propagation(q, gyr, dt)     # Quaternion from previous quaternion and angular velocity
        q = self.oleq(acc, mag, q_g)                    # Second stage: Estimate with OLEQ
        return q



#!/usr/bin/env python3
import unittest
import numpy as np
import ahrs.utils.core as ahrs_core

class TestCoreFunctions(unittest.TestCase):
    def test_assert_same_shapes(self):
        self.assertRaises(ValueError, ahrs_core._assert_same_shapes, [1, 2, 3], [1, 2])

    def test_assert_numerical_iterable(self):
        self.assertRaises(TypeError, ahrs_core._assert_numerical_iterable, [1, 2, 'a'], 'test')

    def test_assert_acc_mag_inputs(self):
        self.assertRaises(ValueError, ahrs_core._assert_acc_mag_inputs, [1, 2, 3], [1, 2])

    def test_get_nan_intervals(self):
        A = np.random.random((10, 3))
        A[[1, 3, 4, 5, 8, 9]] = np.nan
        self.assertEqual(ahrs_core.get_nan_intervals(A), [(1, 1), (3, 5), (8, 9)])

    def test_get_nan_intervals_wrong_input(self):
        A = np.random.random((10, 3))
        A[[1, 3, 4, 5, 8, 9]] = np.nan
        np.testing.assert_array_equal(ahrs_core.get_nan_intervals(A), [(1, 1), (3, 5), (8, 9)])

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
These are the most common routines used to estimate attitude and heading.

"""

from . import mathfuncs
from . import orientation
from . import geometry
from . import frames

from .constants import M_PI, DEG2RAD, RAD2DEG
from .quaternion import Quaternion
from .dcm import DCM



# -*- coding: utf-8 -*-
"""
Attitude obtained with gyroscope and accelerometer-magnetometer measurements,
via complementary filter.

The complementary filter is one of the simplest ways to fuse sensor data from
multiple sensors. It is based on the idea that the errors from one sensor will
be compensated by the other sensor, and vice versa.

The gyroscopes tend to have a low-frequency drift, while the accelerometers
and magnetometers tend to have a high-frequency noise. The complementary filter
simple "combines" these two signals, yielding the benefit of eliminating the
drift from the gyroscope and noise from the accelerometer.

First, the **tilt** angles computed from the accelerometer measurements as:

.. math::
    \\boldsymbol{\\theta}_{am} =
    \\begin{bmatrix}
        \\theta_x \\\\ \\theta_y \\\\ \\theta_z
    \\end{bmatrix} =
    \\begin{bmatrix}
    \\mathrm{arctan2}(a_y, a_z) \\\\
    \\mathrm{arctan2}\\big(-a_x, \\sqrt{a_y^2+a_z^2}\\big) \\\\ 0
    \\end{bmatrix}

Only the roll, :math:`\\theta_x`, and pitch, :math:`\\theta_y`, angles are computed,
leaving the yaw angle, :math:`\\theta_z`, equal to zero.

If a magnetometer sample is available, the **yaw angle**, :math:`\\theta_z` can be
computed. First compensate the measured magnetic field using the tilt:

.. math::
    \\begin{array}{rcl}
    \\mathbf{b} &=&
    \\begin{bmatrix}
        \\cos\\theta_x & \\sin\\theta_x\\sin\\theta_y & \\sin\\theta_x\\cos\\theta_y \\\\
        0 & \\cos\\theta_y & -\\sin\\theta_y \\\\
        -\\sin\\theta_x & \\cos\\theta_x\\sin\\theta_y & \\cos\\theta_x\\cos\\theta_y
    \\end{bmatrix}
    \\begin{bmatrix}m_x \\\\ m_y \\\\ m_z\\end{bmatrix} \\\\
    \\begin{bmatrix}b_x \\\\ b_y \\\\ b_z\\end{bmatrix} &=&
    \\begin{bmatrix}
        m_x\\cos\\theta_x + m_y\\sin\\theta_x\\sin\\theta_y + m_z\\sin\\theta_x\\cos\\theta_y \\\\
        m_y\\cos\\theta_y - m_z\\sin\\theta_y \\\\
        -m_x\\sin\\theta_x + m_y\\cos\\theta_x\\sin\\theta_y + m_z\\cos\\theta_x\\cos\\theta_y
    \\end{bmatrix}
    \\end{array}

Then :math:`\\theta_z` is obtained as:

.. math::
    \\begin{array}{rcl}
    \\theta_z &=& \\mathrm{arctan2}(-b_y, b_x) \\\\
    &=& \\mathrm{arctan2}\\big(m_z\\sin\\theta_y - m_y\\cos\\theta_y, \\; m_x\\cos\\theta_x + \\sin\\theta_x(m_y\\sin\\theta_y + m_z\\cos\\theta_y)\\big)
    \\end{array}

Likewise, the orientation is again estimated at :math:`t`, but this time from a
previous orientation at :math:`t-1`, and a given angular velocity,
:math:`\\boldsymbol{\\omega}=\\begin{bmatrix}\\omega_x & \\omega_y & \\omega_z\\end{bmatrix}^T`,
in rad/s, using the simplest numerical integration:

.. math::
    \\boldsymbol{\\theta}_\\omega =
    \\begin{bmatrix}
    \\theta_{x_t} \\\\ \\theta_{y_t} \\\\ \\theta_{z_t}
    \\end{bmatrix} = 
    \\begin{bmatrix}
    \\theta_{x_{t-1}} + \\omega_x\\Delta_t \\\\
    \\theta_{y_{t-1}} + \\omega_y\\Delta_t \\\\
    \\theta_{z_{t-1}} + \\omega_z\\Delta_t
    \\end{bmatrix}

where :math:`\\Delta_t` is the time interval between the current and previous
measurements, a.k.a. the sampling period or time step. This is merely the
opposite of the sampling frequency: :math:`\\Delta_t=^1/f_s`.

Finally, the estimations are merged using a complementary filter with a
controlling parameter, :math:`\\alpha`, in the range :math:`[0, 1]`:

.. math::
    \\boldsymbol{\\theta} = \\alpha\\boldsymbol{\\theta}_\\omega + (1 - \\alpha)\\boldsymbol{\\theta}_{am}

where :math:`\\boldsymbol{\\theta}_\\omega` is the attitude estimated from the
gyroscope, :math:`\\boldsymbol{\\theta}_{am}` is the attitude estimated from
the accelerometer and the magnetometer, and :math:`\\alpha` is the gain of the
filter.

The filter gain must be a floating value within the range :math:`[0.0, 1.0]`.
It can be seen that when :math:`\\alpha=0`, the attitude is estimated entirely
with the accelerometer and the magnetometer. When :math:`\\alpha=1`, it is
estimated solely with the gyroscope. The values within the range decide how
much of each estimation is "blended" into the quaternion.

"""

import numpy as np
from ..common.quaternion import QuaternionArray
from ..utils.core import _assert_numerical_iterable

class Complementary:
    """
    Complementary filter for attitude estimation as quaternion.

    Parameters
    ----------
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity, in rad/s.
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration, in m/s^2.
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field, in mT.
    frequency : float, default: 100.0
        Sampling frequency in Herz.
    Dt : float, default: 0.01
        Sampling step in seconds. Inverse of sampling frequency. Not required
        if ``frequency`` value is given.
    gain : float, default: 0.9
        Filter gain. Gain equal to 1 uses Angular velocities (gyroscopes) only.
        Gain equal to 0 uses Accelerometer, and Magnetometer if available,
        only. Values greater than zero and less than one blend the two
        estimations proportionally.
    w0 : numpy.ndarray, default: None
        Initial angular position, as roll-pitch-yaw angles, in radians.
    q0 : numpy.ndarray, default: None
        Initial orientation, as a versor (normalized quaternion).
    representation : str, default: ``'angles'``
        Attitude representation. Options are ``'angles'``, ``'quaternion'`` or
        ``'rotmat'``.

    Raises
    ------
    ValueError
        When dimension of input arrays ``acc``, ``gyr``, or ``mag`` are not
        equal.

    """
    def __init__(self,
        gyr: np.ndarray = None,
        acc: np.ndarray = None,
        mag: np.ndarray = None,
        frequency: float = 100.0,
        gain: float = 0.9,
        **kwargs):
        self.gyr: np.ndarray = gyr
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.frequency: float = frequency
        self.Dt: float = kwargs.get('Dt', (1.0/self.frequency) if self.frequency else 0.01)
        self.gain: float = gain
        self.q0: np.ndarray = kwargs.get('q0')
        self.w0: np.ndarray = kwargs.get('w0')
        self._assert_validity_of_inputs()
        # Process of given data
        if self.gyr is not None and self.acc is not None:
            self.W = self._compute_all()

    def _assert_validity_of_inputs(self):
        """Asserts the validity of the inputs."""
        # Assert float values
        for item in ['frequency', 'Dt', 'gain']:
            if isinstance(self.__getattribute__(item), bool):
                raise TypeError(f"Parameter '{item}' must be numeric.")
            if not isinstance(self.__getattribute__(item), (int, float)):
                raise TypeError(f"Parameter '{item}' is not a non-zero number.")
            if item in ['frequency', 'Dt'] and self.__getattribute__(item) <= 0.0:
                raise ValueError(f"Parameter '{item}' must be a non-zero number.")
        if not (0.0 <= self.gain <= 1.0):
            raise ValueError(f"Filter gain must be in the range [0, 1]. Got {self.gain}")
        # Assert arrays
        for item in ['gyr', 'acc', 'mag', 'q0', 'w0']:
            if self.__getattribute__(item) is not None:
                _assert_numerical_iterable(self.__getattribute__(item), item)
                self.__setattr__(item, np.copy(self.__getattribute__(item)))
                if item == 'q0':
                    if self.q0.shape != (4,):
                        raise ValueError(f"Parameter 'q0' must be an array of shape (4,). It is {self.q0.shape}.")
                    if not np.allclose(np.linalg.norm(self.q0), 1.0):
                        raise ValueError(f"Parameter 'q0' must be a versor (norm equal to 1.0). Its norm is equal to {np.linalg.norm(self.q0)}.")
                else:
                    if self.__getattribute__(item).ndim > 2:
                        raise ValueError(f"Input '{item}' must be a one- or two-dimensional array.")
                    array_shape = self.__getattribute__(item).shape
                    if array_shape[-1] != 3:
                        raise ValueError(f"Input '{item}' must be of shape (3,) or (N, 3). Got {array_shape}.")

    def _compute_all(self) -> np.ndarray:
        """
        Estimate the quaternions given all data

        Attributes ``gyr``, ``acc`` and, optionally, ``mag`` must contain data.

        Returns
        -------
        W : numpy.ndarray
            M-by-3 Array with all estimated angles, where M is the number of
            samples.

        """
        if self.gyr.ndim < 2:
            raise ValueError(f"All inputs must have at least two observations.")
        if self.acc.shape != self.gyr.shape:
            raise ValueError(f"Could not operate on acc array of shape {self.acc.shape} and gyr array of shape {self.gyr.shape}.")
        W = np.zeros_like(self.acc)
        if self.mag is None:
            # Estimation with IMU only (Gyroscopes and Accelerometers)
            W2 = self.am_estimation(self.acc)
            W[0] = W2[0] if self.w0 is None else self.w0
        else:
            # Estimation with MARG (IMU and Magnetometer)
            if self.mag.shape != self.gyr.shape:
                raise ValueError(f"Could not operate on mag array of shape {self.mag.shape} and gyr array of shape {self.gyr.shape}.")
            W2 = self.am_estimation(self.acc, self.mag)
            W[0] = W2[0] if self.w0 is None else self.w0
        # Complemetary filter
        if self.mag is None:
            # Estimation with IMU only (Gyroscopes and Accelerometers)
            for i in range(1, len(W)):
                W[i, :2] = (W[i-1, :2] + self.gyr[i, :2]*self.Dt)*self.gain + W2[i, :2]*(1.0-self.gain)
            return W
        # Estimation with MARG (IMU and Magnetometer)
        for i in range(1, len(W)):
            W[i] = (W[i-1] + self.gyr[i]*self.Dt)*self.gain + W2[i]*(1.0-self.gain)
        return W

    def am_estimation(self, acc: np.ndarray, mag: np.ndarray = None) -> np.ndarray:
        """
        Attitude estimation from an Accelerometer-Magnetometer architecture.

        Parameters
        ----------
        acc : numpy.ndarray
            N-by-3 array with measurements of gravitational acceleration.
        mag : numpy.ndarray, default: None
            N-by-3 array with measurements of local geomagnetic field.

        Returns
        -------
        q_am : numpy.ndarray
            Estimated attitude.
        """
        acc = np.copy(acc)
        if acc.ndim < 1:
            raise ValueError(f"Input 'acc' must be a one- or two-dimensional array. Got shape {acc.shape}.")
        if acc.ndim < 2:
            # Estimation with one sample of a tri-axial accelerometer
            a_norm = np.linalg.norm(acc)
            if not a_norm > 0:
                raise ValueError("Gravitational acceleration must be non-zero")
            ax, ay, az = acc/a_norm
            # Tilt from Accelerometer
            ex = np.arctan2( ay, az)                        # Roll
            ey = np.arctan2(-ax, np.sqrt(ay**2 + az**2))    # Pitch
            ez = 0.0                                        # Yaw
            if mag is not None:
                _assert_numerical_iterable(mag, 'Geomagnetic field vector')
                if not(np.linalg.norm(mag) > 0):
                    raise ValueError("Magnetic field must be non-zero")
                mx, my, mz = mag/np.linalg.norm(mag)
                # Get tilted reference frame
                by = my*np.cos(ex) - mz*np.sin(ex)
                bx = mx*np.cos(ey) + np.sin(ey)*(my*np.sin(ex) + mz*np.cos(ex))
                ez = np.arctan2(-by, bx)
            return np.array([ex, ey, ez])
        # Estimation for 2-dimensional arrays
        angles = np.zeros_like(acc)   # Allocation of angles array
        # Estimate tilt angles
        a = acc/np.linalg.norm(acc, axis=1)[:, None]
        angles[:, 0] = np.arctan2(a[:, 1], a[:, 2])
        angles[:, 1] = np.arctan2(-a[:, 0], np.sqrt(a[:, 1]**2 + a[:, 2]**2))
        if mag is not None:
            # Estimate heading angle
            m_norm = np.linalg.norm(mag, axis=1)[:, None]
            if np.where(m_norm == 0)[0].size > 0:
                raise ValueError("All Magnetic field measurements must be non-zero.")
            m = mag/m_norm
            by = m[:, 1]*np.cos(angles[:, 0]) - m[:, 2]*np.sin(angles[:, 0])
            bx = m[:, 0]*np.cos(angles[:, 1]) + np.sin(angles[:, 1])*(m[:, 1]*np.sin(angles[:, 0]) + m[:, 2]*np.cos(angles[:, 0]))
            angles[:, 2] = np.arctan2(-by, bx)
        return angles

    @property
    def Q(self) -> np.ndarray:
        """
        Attitudes as quaternions.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number of
            samples.

        """
        if not hasattr(self, 'W'):
            raise ValueError("No data to perform estimation. Attitude is not computed.")
        return QuaternionArray().from_rpy(self.W)



# -*- coding: utf-8 -*-
"""
The most typical low-cost sensor system is the Accelerometer-Magnetometer
combination (AMC). It integrates the local gravity and the Earth's magnetic
field together, forming a full-attitude estimation system.

The most famous attitude determination formulation is `Wahba's Problem
<https://en.wikipedia.org/wiki/Wahba%27s_problem>`_, which sprung several
solutions using Euler Angles, Direction Cosine Matrices and Quaternions.

The matrix operations in these solutions are the main focus of attention in
this method. The operations are analytically simplified, where the accuracy is
maintained, while the time consumption is reduced, yielding the **Fast
Accelerometer-Magnetometer Combination (FAMC)** :cite:p:`liu2018`, whose main
contributions are:

- Analytic eigenvalue results are given for the dynamic magnetometer reference
  vector.
- Solution as quaternion representation from a simplification of `Davenport's
  q-method <./davenport.html>`_.
- Advantageous on time consumption, compared with existing solutions to Wahba's
  problem.

The AMC relates a Direction Cosine Matrix, :math:`\\mathbf{C}`, such that:

.. math::
    \\begin{array}{c}
    ^b\\mathbf{a} = \\mathbf{C}\\,^r\\mathbf{a} \\\\ \\\\
    ^b\\mathbf{m} = \\mathbf{C}\\,^r\\mathbf{m}
    \\end{array}

where :math:`^b\\mathbf{a}=\\begin{bmatrix}a_x&a_y&a_z\\end{bmatrix}^T` and
:math:`^b\\mathbf{m}=\\begin{bmatrix}m_x&m_y&m_z\\end{bmatrix}^T` are
*normalized* observation vectors from the tri-axial accelerometer and
magnetometer sensors in the body frame :math:`b`, respectively.

The reference frame, :math:`r`, is chosen to follow the `North-East-Down
<https://en.wikipedia.org/wiki/Local_tangent_plane_coordinates#Local_north,_east,_down_(NED)_coordinates>`_
(NED) coordinate system.

Thus, the reference vectors are given by :math:`^r\\mathbf{a}=\\begin{bmatrix}0 & 0 & 1\\end{bmatrix}^T`
and :math:`^r\\mathbf{m}=\\begin{bmatrix}m_N & m_E & m_D\\end{bmatrix}^T`,
although this solution neglects the Easterly geomagnetic field describing it
based on the local `magnetic dip <https://en.wikipedia.org/wiki/Magnetic_dip>`_
:math:`\\theta` to form a simpler expression
:math:`^r\\mathbf{m}=\\begin{bmatrix}\\cos\\theta & 0 & -\\sin\\theta\\end{bmatrix}^T`

The most common solution to Wahba's problem relates two vector observation
pairs with weights. Here, two equivalent weights equal to :math:`0.5` yield the
loss function:

.. math::
    L(\\mathbf{C}) = \\frac{1}{2}\\Big[\\frac{1}{2}\\|^b\\mathbf{a}-\\mathbf{C}\\,^r\\mathbf{a}\\|^2+\\frac{1}{2}\\|^b\\mathbf{m}-\\mathbf{C}\\,^r\\mathbf{m}\\|^2\\Big]

Using `Davenport's q-method <./davenport.html>`_ we find the minimum of
:math:`L(\\mathbf{C})` by calculating the maximum eigenvalue of Davenport's
matrix :math:`\\mathbf{K}` in terms of quaternions:

.. math::
    \\begin{array}{rcl}
    \\mathbf{Kq} &=& \\lambda_{\\mathrm{max}}\\mathbf{q} \\\\
    \\begin{bmatrix}
        \\mathbf{B}+\\mathbf{B}-\\mathrm{tr}(\\mathbf{B})\\mathbf{I} & \\mathbf{z} \\\\
        \\mathbf{z}^T & \\mathrm{tr}(\\mathbf{B})
    \\end{bmatrix}
    \\begin{bmatrix}q_x\\\\q_y\\\\q_z\\\\q_w\\end{bmatrix} &=&
    \\lambda_{\\mathrm{max}}\\begin{bmatrix}q_x\\\\q_y\\\\q_z\\\\q_w\\end{bmatrix}
    \\end{array}

.. warning::
    The definition of the quaternion does NOT follow the common practice of
    setting the scalar part first, but actually putting the vector part first:
    :math:`\\mathbf{q}=\\begin{pmatrix}q_x & q_y & q_z & q_w\\end{pmatrix}`.
    Consider this for all operations. The resulting attitude is, however,
    represented following the common definition at the final step:
    :math:`\\mathbf{q}=\\begin{pmatrix}q_w & q_x & q_y & q_z\\end{pmatrix}`.

The helping arrays are:

.. math::
    \\begin{array}{rcl}
    \\mathbf{B} &=& \\frac{1}{2}\\big(\\,^b\\mathbf{a}(\\,^r\\mathbf{a})^T+\\,^b\\mathbf{m}(^r\\mathbf{m})^T\\big) \\\\ && \\\\
    \\mathbf{z}^T &=& \\begin{bmatrix}B_{23}-B_{32}\\\\B_{31}-B_{13}\\\\B_{12}-B_{21}\\end{bmatrix}
    \\end{array}

.. note::
    Indexing is normally starting from zero, especially in computational setups,
    but the article starts it from one, and it is kept like that in this
    documentation to coincide with the original document.

After knowing that :math:`\\|\\mathbf{a}\\|=\\|\\mathbf{m}\\|=m_N^2+m_D^2=1`,
the four eigenvalues of :math:`\\mathbf{K}` are computed by:

.. math::
    \\begin{array}{rcl}
    \\lambda\\mathbf{K}_1 &=& \\frac{1}{2}\\sqrt{2m_N^2 + 2 + 2m_D^2} = 1 \\\\
    \\lambda\\mathbf{K}_2 &=& \\frac{1}{2}\\sqrt{-2m_N^2 + 2 + 2m_D^2} = \\frac{1}{2}\\sqrt{4m_D^2} = |m_D| \\\\
    \\lambda\\mathbf{K}_3 &=& \\frac{1}{2}\\sqrt{-2m_N^2 + 2 + 2m_D^2} = \\frac{1}{2}\\sqrt{4m_D^2} = -|m_D| \\\\
    \\lambda\\mathbf{K}_4 &=& \\frac{1}{2}\\sqrt{2m_N^2 + 2 + 2m_D^2} = -1
    \\end{array}

The attitude quaternion is the eigenvector of :math:`\\mathbf{K}` corresponding
to the maximum eigenvalue 1. This shows that the AMC is actually
self-constrained and does not require the outer information of the magnetic dip
angle.

.. warning::
    When the AMC is used near the poles there may be two ambiguous solutions
    corresponding to an eigenvalue equal to 1, disorienting the system and
    rendering this estimator useless on polar regions.

Finally, we must compute the eigenvector of the eigenvalule 1. We start
defining:

.. math::
    \\mathbf{S} = \\mathbf{K} - \\mathbf{I}

where :math:`\\mathbf{S}` can be further expanded with matrix row operations in
echelon form :math:`\\mathbf{T}=\\mathbf{\\Lambda}_1\\mathbf{\\Lambda}_2\\mathbf{\\Lambda}_3\\mathbf{S}`:

.. math::
    \\begin{array}{rcl}
    \\mathbf{\\Lambda}_1 &=& \\begin{bmatrix}Y_{11}&&&\\\\Y_{12}&1&&\\\\Y_{13}&&1&\\\\Y_{14}&&&1\\end{bmatrix} \\\\&&\\\\
    \\mathbf{\\Lambda}_2 &=& \\begin{bmatrix}1&Y_{21}&&\\\\&Y_{22}&&\\\\&Y_{23}&1&\\\\&Y_{24}&&1\\end{bmatrix} \\\\&&\\\\
    \\mathbf{\\Lambda}_3 &=& \\begin{bmatrix}1&&Y_{31}&\\\\&1&Y_{32}&\\\\&&Y_{33}&\\\\&&Y_{34}&1\\end{bmatrix}
    \\end{array}

So, :math:`\\mathbf{T}` is expanded as:

.. math::
    \\mathbf{T} = \\begin{bmatrix}1 & & & a \\\\ & 1 & & b \\\\ & & 1 & c \\\\ & & & 0\\end{bmatrix}

where:

.. math::
    \\begin{array}{rcl}
    a &=& B_{23}\\big[Y_{11}+Y_{12}(Y_{21}+Y_{23}Y_{31})+Y_{13}Y_{31}\\big] - (B_{13}-B_{31})(Y_{21}+Y_{23}Y_{31}) - Y_{31}B_{21} \\\\
    b &=& B_{23}\\big[Y_{12}(Y_{22}+Y_{23}Y_{32})+Y_{13}Y_{32}\\big] - (B_{13}-B_{31})(Y_{22}+Y_{23}Y_{32}) - Y_{32}B_{21} \\\\
    c &=& B_{23}(Y_{13}Y_{33}+Y_{12}Y_{23}Y_{33}) - Y_{33}B_{21} - Y_{23}Y_{33}(B_{13}-B_{31})
    \\end{array}

The *unnormalized scalar-wise* quaternion is given by the solution:

.. math::
    \\mathbf{q} = \\begin{pmatrix}-1 & a & b & c\\end{pmatrix}

which is easily normalized:

.. math::
    \\mathbf{q} = \\frac{1}{\\sqrt{a^2+b^2+c^2+1}}\\begin{pmatrix}-1 & a & b & c\\end{pmatrix}

"""

import numpy as np

from ..utils.core import _assert_acc_mag_inputs

class FAMC:
    """
    Fast Accelerometer-Magnetometer Combination

    Parameters
    ----------
    acc : numpy.ndarray, default: None
        M-by-3 array with measurements of acceleration in in m/s^2
    mag : numpy.ndarray, default: None
        M-by-3 array with measurements of magnetic field in mT

    Attributes
    ----------
    acc : numpy.ndarray
        M-by-3 array with M accelerometer samples.
    mag : numpy.ndarray
        M-by-3 array with m magnetometer samples.
    Q : numpy.array, default: None
        M-by-4 Array with all estimated quaternions, where M is the number of
        samples. Equal to None when no estimation is performed.

    Raises
    ------
    ValueError
        When dimension of input arrays ``acc`` and ``mag`` are not equal.

    Examples
    --------
    >>> acc_data.shape, mag_data.shape      # NumPy arrays with sensor data
    ((1000, 3), (1000, 3))
    >>> from ahrs.filters import FAMC
    >>> famc = FAMC(acc=acc_data, mag=mag_data)
    >>> famc.Q       # Estimated attitudes as Quaternions
    array([[-0.82311077,  0.45760535, -0.33408929, -0.0383452 ],
           [-0.82522048,  0.4547043 , -0.33277675, -0.03892033],
           [-0.82463698,  0.4546915 , -0.33422422, -0.03903417],
           ...,
           [-0.82420642,  0.56217735,  0.02548005, -0.06317571],
           [-0.82364606,  0.56311099,  0.0241655 , -0.06268338],
           [-0.81844766,  0.57077781,  0.02532182, -0.06095017]])
    >>> famc.Q.shape
    (1000, 4)

    """
    def __init__(self, acc: np.ndarray = None, mag: np.ndarray = None) -> None:
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.Q: np.ndarray = None
        if self.acc is not None and self.mag is not None:
            self.Q = self._compute_all()

    def _compute_all(self) -> np.ndarray:
        """
        Estimate the quaternions given all data.

        Attributes ``acc`` and ``mag`` must contain data.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        _assert_acc_mag_inputs(self.acc, self.mag)
        acc, mag = np.copy(self.acc), np.copy(self.mag)
        if acc.ndim < 2:
            return self.estimate(acc, mag)
        num_samples = len(acc)
        return np.array([self.estimate(acc[t], mag[t]) for t in range(num_samples)])

    def estimate(self, acc: np.ndarray, mag: np.ndarray) -> np.ndarray:
        """
        Attitude Estimation

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer.
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion of the form :math:`\\begin{pmatrix}q_w & q_x & q_y & q_z\\end{pmatrix}`

        Examples
        --------
        >>> acc_data = np.array([4.098297, 8.663757, 2.1355896])
        >>> mag_data = np.array([-28.71550512, -25.92743566, 4.75683931])
        >>> from ahrs.filters import FAMC
        >>> famc = FAMC()
        >>> famc.estimate(acc=acc_data, mag=mag_data)   # Estimate attitude as quaternion
        array([-0.82311077,  0.45760535, -0.33408929, -0.0383452])

        """
        # Normalize measurements (eq. 10)
        a_norm = np.linalg.norm(acc)
        m_norm = np.linalg.norm(mag)
        if not a_norm > 0 or not m_norm > 0:    # handle NaN
            return None
        ax, ay, az = acc / a_norm               # A = [ax, ay, az] in body frame
        mx, my, mz = mag / m_norm               # M = [mx, my, mz] in body frame
        # Dynamic magnetometer reference vector
        m_D = ax*mx + ay*my + az*mz             # (eq. 13)
        m_N = np.sqrt(1.0 - m_D**2)
        # Parameters
        B = np.zeros((3, 3))                    # (eq. 18)
        B[:, 0] = m_N * np.array([mx, my, mz])
        B[:, 2] = m_D * np.array([mx, my, mz]) + np.array([ax, ay, az])
        B *= 0.5
        tau = B[0, 2] + B[2, 0]
        alpha = np.zeros(3)
        Y = np.zeros((3, 3))
        # First Row
        alpha[0] = B[2, 2] - B[0, 0] + 1.0
        Y[0, 0] = -1.0 / alpha[0]
        Y[0, 1] = B[1, 0] / alpha[0]
        Y[0, 2] = tau / alpha[0]
        # Second Row
        alpha[1] = -B[1, 0]**2/alpha[0] + B[0, 0] + B[2, 2] + 1
        Y[1, 0] = -B[1, 0] / (alpha[0] * alpha[1])
        Y[1, 1] = -1.0 / alpha[1]
        Y[1, 2] = (B[1, 2] + B[1, 0]*tau/alpha[0]) / alpha[1]
        # Third row
        alpha[2] = alpha[0] - 2 + tau**2/alpha[0] + Y[1, 2]**2*alpha[1]
        Y[2, 0] = (tau/alpha[0] + B[1, 0]*Y[1, 2]/alpha[0]) / alpha[2]
        Y[2, 1] = Y[1, 2] / alpha[2]
        Y[2, 2] = 1.0 / alpha[2]
        # Quaternion Elements (eq. 21)
        a = B[1, 2]*(Y[0, 0] + Y[0, 1]*(Y[1, 2]*Y[2, 0] + Y[1, 0]) + Y[0, 2]*Y[2, 0]) - (B[0, 2]-B[2, 0])*(Y[1, 2]*Y[2, 0] + Y[1, 0]) - Y[2, 0]*B[1, 0]
        b = B[1, 2]*(          Y[0, 1]*(Y[1, 2]*Y[2, 1] + Y[1, 1]) + Y[0, 2]*Y[2, 1]) - (B[0, 2]-B[2, 0])*(Y[1, 2]*Y[2, 1] + Y[1, 1]) - Y[2, 1]*B[1, 0]
        c = B[1, 2]*(          Y[0, 1]* Y[1, 2]*Y[2, 2]            + Y[0, 2]*Y[2, 2]) - (B[0, 2]-B[2, 0])*(Y[1, 2]*Y[2, 2])           - Y[2, 2]*B[1, 0]
        q = np.array([-1, a, b, c])         # (eq. 22)
        return q/np.linalg.norm(q)          # (eq. 23)
